import logging

import h5py
import numpy as np

import radiotools.coordinatesystems
import radiotools.atmosphere.models as models

from radiotools import helper as hp
from scipy.constants import c as c_vacuum

from ..utilities import angle_between, bandpass_filter_trace
from smiet import units


class Shower:
    """
    The ``Shower`` class represents an air shower setup and evolution.

    It holds the geometric information for a shower, like the zenith/azimuth, as well as the longitudinal profile.
    Next to this, it can also store the magnetic field, the atmosphere in which it was simulated, and the core.
    These variables are used to calculate the `radiotools.coordinatesystems.cstrafo` object, which can be used
    to convert between different coordinate systems.
    The primary use case is to specify the target parameters for a shower in template synthesis.

    Attributes
    ----------
    atmosphere : radiotools.atmosphere.models.Atmosphere
        The atmosphere in which the shower was simulated
    core : np.ndarray
        The position of the core in the NRR coordinate system
    """

    def __init__(self, atm_model=17):
        self.logger = logging.getLogger("smiet.numpy.io.Shower")

        self._grammages = None
        self._long_profile = None
        self._zenith = None
        self._azimuth = None
        self._magnetic_field = None
        self._slice_grammage = None

        self.atmosphere = models.Atmosphere(
            model = atm_model,
            curved=False
        )
        self.core = [0, 0, 0]

        # To be computed
        self._GH_parameters = None
        self._transformer = None

    @property
    def xmax(self):
        """
        The :math:`X_{max}` of the shower. This is the result from fitting a GH to the longitudinal profile (not set manually).
        """
        if self._GH_parameters is not None:
            return self._GH_parameters["xmax"]
        else:
            self.logger.error(
                "The Gaisser-Hillas parameters have not been set. You probably did not set the longitudinal profile yet."
            )

    @property
    def nmax(self):
        """
        The :math:`N_{max}` of the shower. This results from fitting a GH to the longitudinal profile (ie not set manually).
        """
        if self._GH_parameters is not None:
            return self._GH_parameters["nmax"]
        else:
            self.logger.error(
                "The Gaisser-Hillas parameters have not been set. You probably did not set the longitudinal profile yet."
            )

    @property
    def magnet(self):
        """
        Magnetic field vector in the NRR coordinate system
        """
        if self._magnetic_field is not None:
            return self._magnetic_field
        else:
            self.logger.error("The magnetic field vector has not been set")

    @magnet.setter
    def magnet(self, magnet_field_vec):
        assert len(magnet_field_vec) == 3, (
            "B-field vector must contain three components"
        )

        self._magnetic_field = magnet_field_vec
        self._transformer = None  # reset the transformer to be recalculated

    @property
    def grammages(self):
        """
        The grammages at which the longitudinal profile is defined in.
        """
        if self._grammages is not None:
            return self._grammages
        else:
            self.logger.error("Grammages not yet set.")

    @grammages.setter
    def grammages(self, grams: np.ndarray):
        self._grammages = grams

    @property
    def long_profile(self):
        """
        The longitudinal profile of the shower, containing the sum of positrons and electrons. The profile should have the shape (slices,).
        """
        if self._long_profile is not None:
            return self._long_profile
        else:
            self.logger.error("The longitudinal profile has not been set")

    @long_profile.setter
    def long_profile(self, profile: np.ndarray):
        self._long_profile = profile
        # self._GH_parameters = self.fit_gaisser_hillas()

    @property
    def slice_grammage(self):
        """
        The (slant depth) step width with which the longitudinal profile is sampled.
        This is inferred from the longitudinal profile when that one is set.
        """
        if self._grammages is not None:
            return self._grammages[1] - self._grammages[0]
        else:
            self.logger.error(
                "Slice thickness not yet set. Copy it from another shower or set a long profile."
            )

    @property
    def zenith(self):
        """
        The zenith angle of the shower (in internal units).
        """
        if self._zenith is not None:
            return self._zenith
        else:
            self.logger.error("Geometry has not been set")

    @zenith.setter
    def zenith(self, zenith: float) -> None:
        self._zenith = zenith
        self._transformer = None  # reset the transformer to be recalculated

    @property
    def azimuth(self):
        """
        The azimuth angle of the shower (in internal units).
        """
        if self._azimuth is not None:
            return self._azimuth
        else:
            self.logger.error("Geometry has not been set")

    @azimuth.setter
    def azimuth(self, azimuth: float) -> None:
        self._azimuth = azimuth
        self._transformer = None  # reset the transformer to be recalculated

    @property
    def geomagnetic_angle(self):
        """
        Calculate the angle between the magnetic field vector and the shower axis
        """
        shower_axis = hp.spherical_to_cartesian(
            self.zenith / units.rad, self.azimuth / units.rad
        )

        return angle_between(self.magnet * -1, shower_axis)

    @property
    def GH_parameters(self) -> dict:
        """Get the Gaisser-Hillas parameters as a dictionary."""
        if self._GH_parameters is not None:
            return self._GH_parameters
        else:
            self.logger.error("GH parameters have not been set")

    @GH_parameters.setter
    def GH_parameters(self, params: dict) -> None:
        self._GH_parameters = params

    def fit_gaisser_hillas(self):
        """
        Fit a Gaisser-Hillas function to the currently loaded longitudinal profile,
        to determine the parameters like :math:`X_{max}` and :math:`N_{max}`.

        Returns
        -------
        popt : np.ndarray
            The parameters of the Gaisser-Hillas function, in the order [N, X0, Xmax, p0, p1, p2]
            (p0,p1,p2 are the polynomial coefficients in denominator for lambda)

        Notes
        -----
        The Gaisser-Hillas function is defined as:

        .. math::
            f(X) = N \\left( \\frac{X - X_0}{X_{max} - X_0} \\right)^{\\frac{X_{max} - X_0}{\\lambda}} \\exp\\left( \\frac{X_{max} - X}{\\lambda} \\right)
        """
        from scipy.optimize import curve_fit

        def gaisser_hillas(X, N, X0, Xmax, p0, p1=0, p2=0):
            l = p0 + p1 * X + p2 * X**2
            power = (Xmax - X0) / l

            if np.sum(l < 0):
                return np.inf

            if np.sum(power < 0):
                return np.inf

            if np.sum(power > 100):
                return np.inf

            result = np.zeros_like(X)
            mask = (X - X0) >= 0
            result[mask] = (
                N
                * ((X[mask] - X0) / (Xmax - X0)) ** (power[mask])
                * np.exp((Xmax - X[mask]) / l[mask])
            )
            result = np.nan_to_num(result)

            return result

        popt = curve_fit(
            gaisser_hillas,
            self._grammages,
            self.long_profile,
            p0=[
                self.long_profile.max(),
                0,
                self._grammages[self.long_profile.argmax()],
                20,
            ],
            maxfev=3000,
        )[0]

        # calculate L and R from X0 and lambda
        L = np.sqrt(
            np.abs((popt[2] - popt[1]) * (popt[3]))
        )  # L  = sqrt|(Xmax - X0)*lambda|
        R = np.sqrt(
            np.abs(popt[3] / (popt[2] - popt[1]))
        )  # R = sqrt|lambda/(Xmax - X0)|

        return {
            "nmax": popt[0],
            "x0": popt[1],
            "xmax": popt[2],
            "lambda": popt[3],
            "L": L,
            "R": R,
        }

    def get_transformer(self):
        """
        Get the transformer object from radiotools which can be used to convert between different coordinate systems.
        The transformer is cached, so that it is only calculated once. If the geometry or magnetic field vector gets
        updated, the transformer is reset.

        Returns
        -------
        transformer : radiotools.coordinatesystems.cstrafo
            The cstrafo object with the current geometry and magnetic field vector.
        """
        if self._transformer is None:
            self._transformer = radiotools.coordinatesystems.cstrafo(
                self.zenith / units.rad,
                self.azimuth / units.rad,
                magnetic_field_vector=self._magnetic_field,  # units of magnet don't matter, only direction
            )
        return self._transformer

    def copy_settings(self, other):
        """
        This convenience method allows to easily copy the settings from another Shower,
        which is useful when for example creating a target Shower.

        Parameters
        ----------
        other: Shower
            The Shower to copy the settings from.

        Notes
        -----
        The settings which are copied from the other shower are:

        * The geometry
        * The magnetic field vector
        * The simulation core
        * The slice grammage
        * The atmosphere
        """
        self.zenith = other.zenith
        self.azimuth = other.azimuth
        self.magnet = other.magnet
        self.core = other.core
        self.atmosphere = other.atmosphere

        self._slice_grammage = other.slice_grammage

    def set_parameters(
        self,
        grammages,
        long_profile,
        xmax,
        nmax,
        zenith,
        azimuth,
        magnetic_field,
        core,
    ) -> None:
        """
        Set the parameters of the shower model from a dictionary of parameters.

        This is a convenience function to set all parameters at once.

        Parameters
        ----------
        grammages : np.ndarray
            an array of atmospheric depth in g/cm^2
        long_profile : np.ndarray
            an array of the longitudinal profile of the shower
        xmax : float
            depth of shower maximum in g/cm^2. Not used for now as the 
            values are directly fitted from the profile.
        nmax : float
            number of particles at shower maximum.  Not used for now as the 
            values are directly fitted from the profile.
        zenith : float
            zenith angle in radians
        azimuth : float
            azimuthal angle in radians
        magnetic_field : np.ndarray
            magnetic field vector in Gauss. Must be a 3-vector
        core : np.ndarray
            core position in meters. Must be a 3-vector
        """
        self.grammages = grammages
        self.long_profile = long_profile

        self.zenith = zenith
        self.azimuth = azimuth
        self.magnet = magnetic_field
        self.core = core

        self.GH_parameters = self.fit_gaisser_hillas()

    def transform_profile_to_origin(self, origin_grammages) -> None:
        """
        Transform the longitudinal profile to match those of the origin shower.

        This is needed such that the grammages of the origin and target match.

        Parameters
        ----------
        origin_grammages : jax.typing.ArrayLike
            The grammages of the origin shower.
        """
        self.long_profile = np.interp(
            x=origin_grammages, xp=self.grammages, fp=self.long_profile
        )
        self.grammages = origin_grammages



class CoreasHDF5:
    """
    The ``CoreasHDF5`` class is a base class for reading in HDF5 files produced by CoREAS.

    It extracts the settings used in the CoREAS simulation, like the time resolution and the core position.
    It also reads in settings from the CORSIKA part, such as the geometry and the magnetic field.
    These two are available as dictionaries in the ``coreas_settings`` and ``corsika_settings`` properties.
    Next to the settings, it is also possible to get in the longitudinal profile of the shower.
    Furthermore, the class offers a method to read in the time traces of an observer in the HDF5 file.

    Note that it returns all quantities in the NRR coordinate system, which is different from
    the CORSIKA coordinate system.

    Parameters
    ----------
    file_path : str
        The path to the CoREAS HDF5 file
    """

    def __init__(self, file_path):
        self._file = file_path
        self._coreas_settings = None
        self._corsika_settings = None
        self._trace_length = None

        self.logger = logging.getLogger("smiet.numpy.io.CoreasHDF5")

    @property
    def trace_length(self):
        """
        The number of samples in each time trace. This is calculated from the time boundaries
        (automatic or manual) and the time resolution of the simulation.
        """
        if self._trace_length is None:
            trace_length = int(
                self.coreas_settings["automatic_time_boundaries"]
                / self.coreas_settings["time_resolution"]
            )
            if trace_length != 0:
                self._trace_length = trace_length
            else:  # if automatic time boundaries are not set, use manual time boundaries
                self._trace_length = int(
                    self.coreas_settings["manual_time_boundaries"]
                    / self.coreas_settings["time_resolution"]
                )
        return self._trace_length

    @property
    def coreas_settings(self):
        """
        Get the CoREAS configuration of this simulation. This is contained in a dictionary, with the following
        keys:

        * core_coordinate: The core position in the NRR coordinate system
        * time_resolution: The time resolution of the simulation
        * automatic_time_boundaries: The automatic time boundaries of the simulation
        * manual_time_boundaries: The manual time boundaries of the simulation
        * n0: The refractive index at sea level
        """
        if self._coreas_settings is None:
            self._coreas_settings = self.__get_coreas_settings()
        return self._coreas_settings

    @property
    def corsika_settings(self):
        """
        Get the CORSIKA configuration of this simulation. This is contained in a dictionary, with the following
        keys:

        * zenith: The zenith angle of the shower
        * azimuth: The azimuth angle of the shower
        * magnetic_field: The magnetic field vector in the NRR coordinate system
        * observation_level: The observation level of the shower
        * atmosphere_model: The atmosphere model used in the simulation
        * primary_particle: The primary particle used in the simulation
        * primary_energy: The energy of the primary particle
        * thinning_level: The thinning level used in the simulation
        """
        if self._corsika_settings is None:
            self._corsika_settings = self.__get_corsika_settings()
        return self._corsika_settings

    @property
    def name(self):
        """
        Get the name of the simulation, which is the filename without the extension.
        """
        return self._file.split("/")[-1].split(".")[0]

    def get_long_profile(self):
        """
        Get the longitudinal profile of the shower from the HDF5 file stored in the `self._file` attribute.

        The profile is read from the "NumberOfParticles" dataset, which sits in the "atmosphere" group.

        Returns
        -------
        grammages: np.ndarray
            The grammages at which the longitudinal profile is defined
        long_profile: np.ndarray
            The longitudinal profile of the shower
        """
        with h5py.File(self._file) as file:
            grammages = file["atmosphere"]["NumberOfParticles"][:, 0]
            long_profile = file["atmosphere"]["NumberOfParticles"][:, 2:4].sum(axis=1)
        return grammages, long_profile

    def get_gaisser_hillas_parameters(self):
        """
        Get the Gaisser-Hillas parameters of the shower from the HDF5 file stored in the `self._file` attribute.

        The parameters are read from the "GaisserHillasParameters" dataset, which sits in the "atmosphere" group.

        Returns
        -------
        gh_parameters: dict
            A dictionary containing the Gaisser-Hillas parameters of the shower.
            The keys are: 'nmax', 'x0', 'xmax', 'lambda', 'L', 'R'
        """
        with h5py.File(self._file) as file:
            GH_param = file["atmosphere"].attrs["Gaisser-Hillas-Fit"]
        lmbda = np.mean(
                GH_param[3]
                + GH_param[4] * self.grammages
                + GH_param[5] * self.grammages**2)
        
        GH_parameters = {
            "nmax": GH_param[0],
            "X0": GH_param[1],
            "xmax": GH_param[2],
            "lambda": lmbda,
            "L" : np.sqrt(np.abs((GH_param[2] - GH_param[1]) * lmbda)),
            "R" : np.sqrt(np.abs(lmbda / (GH_param[2] - GH_param[1])))
        }

        return GH_parameters

    def filter_trace(self, trace, f_min, f_max):
        """
        Filter a trace between `f_min` and `f_max` using the internal time resolution of the simulation.

        The function assumes the time samples are on the last axis, but if this dimension does not match
        the trace length it will check the other axes to see if any other dimension matches the trace length.
        The filter used is a straightforward bandpass filter, using the :func:`bandpass_filter_trace` function.

        Parameters
        ----------
        trace: np.ndarray
            The trace to filter. Can be multidimensional.
        f_min: float
            The lower frequency (in internal units)
        f_max: float
            The upper frequency (in internal units)

        Returns
        -------
        filtered_trace: np.ndarray
            The bandpass filtered trace
        """
        trace_axis = -1  # based on self.get_trace()
        if trace.shape[trace_axis] != self.trace_length:
            self.logger.warning(
                "Trace shape does not match recorded trace length along the last axis"
            )
            self.logger.info("Attempting to find the trace axis...")
            for shape_i in range(len(trace.shape)):
                if trace.shape[shape_i] == self.trace_length:
                    self.logger.info(
                        f"Found axis {shape_i} which matches trace length!"
                    )
                    trace_axis = shape_i
                    break
        return bandpass_filter_trace(
            trace,
            self.coreas_settings["time_resolution"],
            f_min,
            f_max,
            sample_axis=trace_axis,
        )

    def __get_coreas_settings(self):
        with h5py.File(self._file) as file:
            time_resolution = file["CoREAS"].attrs["TimeResolution"] * units.s
            auto_trace_time_length = (
                file["CoREAS"].attrs["AutomaticTimeBoundaries"] * units.s
            )
            trace_time_length = (
                file["CoREAS"].attrs["TimeUpperBoundary"]
                - file["CoREAS"].attrs["TimeLowerBoundary"]
            ) * units.s
            core = (
                np.array(
                    [
                        -1 * file["CoREAS"].attrs["CoreCoordinateWest"],
                        file["CoREAS"].attrs["CoreCoordinateNorth"],
                        file["CoREAS"].attrs["CoreCoordinateVertical"],
                    ]
                )
                * units.cm
            )
            n0 = file["CoREAS"].attrs["GroundLevelRefractiveIndex"]

        return {
            "core_coordinate": core,
            "time_resolution": time_resolution,
            "automatic_time_boundaries": auto_trace_time_length,
            "manual_time_boundaries": trace_time_length,
            "n0": n0,
        }

    def __get_corsika_settings(self):
        with h5py.File(self._file) as file:
            zenith = file["inputs"].attrs["THETAP"][0] * units.deg
            azimuth = (
                file["inputs"].attrs["PHIP"][0] * units.deg - 90 * units.deg
            )  # transform to radiotools coord
            magnet = (
                np.array(
                    [
                        0,
                        file["inputs"].attrs["MAGNET"][0],
                        -1 * file["inputs"].attrs["MAGNET"][1],
                    ]
                )
                * units.microtesla
            )
            obs_lev = file["inputs"].attrs["OBSLEV"] * units.m
            atm_model = file["inputs"].attrs["ATMOD"]
            primary = file["inputs"].attrs["PRMPAR"]
            energy = file["inputs"].attrs["ERANGE"][0] * units.GeV
            thinning = file["inputs"].attrs["THIN"][0]

        return {
            "zenith": zenith,
            "azimuth": azimuth,
            "magnetic_field": magnet,
            "observation_level": obs_lev,
            "atmosphere_model": atm_model,
            "primary_particle": primary,
            "primary_energy": energy,
            "thinning_level": thinning,
        }

    def get_trace_raw(self, observer_name):
        """
        Get the raw trace (as produced by CoREAS) from an observer in the HDF5.

        The trace is converted to the internal coordinate system before being returned.

        Parameters
        ----------
        observer_name : str
            The name of the observer as defined in the HDF5 file

        Returns
        -------
        trace_slice_ground : np.ndarray
            The raw trace (on the ground), shaped as (trace_length, 3).
        trace_time_array : np.ndarray
            The array containing the timings of the samples, shaped as (trace_length,).
            This is just the first column of the CoREAS output, converted to internal units.
        """
        with h5py.File(self._file) as file:
            try:
                trace_slice = file["CoREAS"]["observers"][f"{observer_name}"][
                    :
                ]  # samples x 4
                self.logger.debug(
                    f"Successfully read in trace of observer {observer_name}"
                )
            except KeyError:
                raise ValueError(f"Observer {observer_name} not present in the file")

            # Save trace start time
            trace_time_array = trace_slice[:, 0] * units.s

            # Convert CGS to internal units
            trace_slice *= c_vacuum * 1e2 * units.microvolt / units.m
            trace_slice_ground = np.array(
                [-trace_slice[:, 2], trace_slice[:, 1], trace_slice[:, 3]]
            )

        return trace_slice_ground, trace_time_array



"""
Main class to run SMIET on
"""

import os
import logging
import numpy as np
from typing_extensions import Self, Union

import smiet.units as units
import smiet.numpy as smiet_np

try:
    import jax
    import jax.numpy as jnp
    import smiet.jax as smiet_jax
except ImportError:
    jax = None
    jnp = None
    smiet_jax = None


class SMIETRun:
    """
    Main class to run SMIET on. The idea of this class is so that all users can easily apply SMIET to their own analyses.

    Main Features:
    --------------
    - generating templates from origin showers, in particular from generated origin shower libraries
    - loading templates from existing template libraries
    - loading target showers from existing Coreas simulations
    - generating target showers from user-provided longitudinal profiles
    - synthesizing target showers from origin shower templates + target showers, with interpolation (by default)
    - retrieving time traces, frequency spectra, and fluences for synthesized showers

    The user is expected to give the site in which SMIET will apply on, and the backend (jax or numpy).
    """

    __site_parameters = {
        "base": {
            "magnetic_field_vector": np.array(
                [0.004675, 0.186270, -0.456412]
            ),  # in Gauss
            "frequency_range": [30, 500, 100],  # in MHz
        },
        "lofar": {
            "magnetic_field_vector": np.array(
                [0.004675, 0.186270, -0.456412]
            ),  # in Gauss
            "frequency_range": [30, 80, 50],  # in MHz
        },
        "ska": {
            "magnetic_field_vector": np.array([0.0, 2.760, 4.827]),  # in Gauss
            "frequency_range": [30, 500, 100],  # in MHz
        },
        "auger": {
            "magnetic_field_vector": np.array(
                [0.00871198, 0.19693423, 0.1413841]
            ),  # in Gauss
            "frequency_range": [30, 80, 50],  # in MHz
        },
    }

    def __init__(
        self: Self,
        site: str = "base",
        backend: str = "numpy",
    ) -> None:
        """
        Initialize the SMIET runner.

        This will set up default parameters and configurations.

        Parameters:
        ----------
        site : str, default = "base"
            The site to run the simulation for. Default is "base", which uses the LOFAR magnetic field with maximal frequency range ([30, 500] MHz), which are the values which SMIET is tested on.

            Current available options are:
            - "base": LOFAR magnetic field with maximal frequency range ([30, 500] MHz)
            - "lofar": LOFAR site parameters (magnetic field and frequency range [30, 80] MHz)
            - "ska": SKA site parameters (magnetic field and frequency range [30, 500] MHz)
            - "auger": Pierre Auger Observatory (AERA) site parameters (magnetic field and frequency range [30, 80] MHz)
        backend : str
            The backend to use for computations. Options are "numpy", "jax_cpu", or "jax_cuda". Default is "numpy".

            If "jax_cpu" or "jax_cuda" is selected but JAX is not installed, it will fall back to "numpy".
            In this case, a warning to install SMIET with the JAX backend will be printed.

            Here, "jax_cpu" will use JAX with CPU as the device, while "jax_cuda" will use JAX with GPU (CUDA) as the device, if available.
            If a GPU is not available, it will fall back to CPU.

        Returns:
        --------
        None
        """
        self.site = site
        self.__get_site_parameters(site)
        self.logger = logging.getLogger("smiet.core.SMIETRun")

        if backend == "numpy":
            self.backend = "numpy"
            self.smiet = smiet_np
        elif backend.find("jax") != -1:
            raise NotImplementedError(
                "JAX backend is not yet implemented in this version of SMIET. Please use the numpy backend for now."
            )
            if jax is None or smiet_jax is None:
                self.logger.warning(
                    "JAX backend selected but JAX is not installed. Falling back to numpy backend. To use JAX, please install SMIET with the JAX backend."
                )
                self.backend = "numpy"
                self.smiet = smiet_np
            else:
                self.backend = "jax"
                if backend == "jax_cpu":
                    jax.config.update("jax_default_device", jax.devices("cpu")[0])
                self.smiet = smiet_jax

        # some flags
        self.__loaded_templates = False
        self.__loaded_targets = False
        self.__synthesized_showers = False

        # containers for origins
        self.origin_xmaxs = None
        self.origin_zeniths = None
        self.syntheses = []

        # containers for targets
        self.target_xmaxs = None
        self.target_zeniths = None
        self.targets = []

        # containers for synthesis
        self.synth_traces = []
        self.synth_times = []

        # additional properties that are useful for later
        self.antenna_positions = []
        self.antenna_labels = None
        self.delta_t = None

    def __get_site_parameters(self: Self, site: str) -> None:
        """
        Retrieve the parameters related to each site.

        This includes the magnetic field and frequency range.

        Parameters:
        ----------
        site : str
            The site to retrieve parameters for.
        """
        self.freq_ar = self.__site_parameters[site]["frequency_range"]
        self.magnet = self.__site_parameters[site]["magnetic_field_vector"]

    def generate_templates(
        self: Self,
        origin_shower_repo: str,
        template_repo: str,
    ) -> None:
        """
        Generate templates from a list of origin showers generated from the origin shower library.

        This is meant to be ran once to generate the templates used for the origin shower library.

        Parameters:
        ----------
        origin_shower_repo : str
            The path to the origin shower repository containing all the origin showers to generate templates from.
        template_repo : str
            The path to the repository to save the generated templates.
        """
        # first check if the template repo exists
        if not os.path.exists(template_repo):
            raise FileNotFoundError(
                f"Template repository {template_repo} does not exist."
            )

        # get the path to all the templates
        path_to_templates = os.path.join(template_repo, self.site, self.backend)

        # read the sim_labels.txt file and extract the info for the configurations
        # this loads the following columns:
        # SIM_ID  XMAX  ZENITH  ITERATION
        config_data = np.genfromtxt(
            os.path.join(origin_shower_repo, "sim_labels.txt"),
            delimiter="\t",
            skip_header=1,
        )

        # get an estimate of how long it will take
        n_showers = len(
            [f for f in os.listdir(origin_shower_repo) if f.endswith(".hdf5")]
        )
        self.logger.warning(
            f"Generating {n_showers} templates from origin showers. This will take at most {n_showers * 10 // 60} minutes."
        )

        # for each origin shower, make a template and save it
        config_txt = "xmax,zenith\n"
        for origin_shower_file in os.listdir(origin_shower_repo):
            if not origin_shower_file.endswith(".hdf5"):
                continue

            # add sim id identification
            # find the corresponding row in the config file
            origin_shower_sim_id = origin_shower_file.replace("SIM", "").replace(
                ".hdf5", ""
            )

            sim_row_idx = np.argwhere(config_data[:, 0] == int(origin_shower_sim_id))[
                0
            ][0]
            # get the properties
            zenith = int(config_data[sim_row_idx, 1])
            xmax = int(config_data[sim_row_idx, 2])
            iter_num = int(config_data[sim_row_idx, 3])

            config_txt += f"{xmax},{zenith}\n"

            origin_shower = self.smiet.SlicedShower(
                os.path.join(origin_shower_repo, origin_shower_file)
            )

            # get a label for the directory based on xmax and zenith
            save_path = os.path.join(path_to_templates, f"xmax{xmax}_zen{zenith}")
            template_name = f"template_sim{iter_num}.hdf5"

            # check if it exists already
            if os.path.exists(os.path.join(save_path, template_name)):
                self.logger.info(
                    f"Template for origin shower {origin_shower_file} already exists. Skipping."
                )
                continue

            # make if it doesnt exist already
            if not os.path.exists(save_path):
                os.makedirs(save_path, exist_ok=True)

            synthesis = self.smiet.TemplateSynthesis(freq_ar=self.freq_ar)

            synthesis.make_template(origin_shower)

            # now save the template based on the iteration number
            synthesis.save_template(
                save_dir=save_path, template_file=f"template_sim{iter_num}.hdf5"
            )

            if self.backend == "jax":
                # clear cache everytime to avoid memory issues
                jax.clear_caches()

        # finally for each configuration, get the xmax value and set the closest one to
        # the true value to be sim1
        for xmax_zenith in os.listdir(path_to_templates):
            config_path = os.path.join(path_to_templates, xmax_zenith)
            if not os.path.isdir(config_path):
                continue
            xmax_diff = []
            for config_file in os.listdir(config_path):
                if not config_file.endswith(".hdf5"):
                    continue
                template_path = os.path.join(config_path, config_file)
                synthesis = self.smiet.TemplateSynthesis(freq_ar=self.freq_ar)
                synthesis.load_template(template_path)

                # check if this is the closest to the true xmax
                true_xmax = float(xmax_zenith.split("_")[0].replace("xmax", ""))
                synth_xmax = synthesis.template_information["xmax"]
                xmax_diff.append(np.abs(true_xmax - synth_xmax))
            # get the index of the closest one
            closest_index = np.argmin(xmax_diff)
            # rename the file to sim1
            closest_file = os.listdir(config_path)[closest_index]
            if closest_file != "template_sim1.hdf5":
                # rename the the original sim1 file to something temporary first
                os.rename(
                    os.path.join(config_path, "template_sim1.hdf5"),
                    os.path.join(config_path, "temp.hdf5"),
                )
                # then rename the closest file to sim 1
                os.rename(
                    os.path.join(config_path, closest_file),
                    os.path.join(config_path, "template_sim1.hdf5"),
                )
                # finally replace temp file to the closest file name
                os.rename(
                    os.path.join(config_path, "temp.hdf5"),
                    os.path.join(config_path, closest_file),
                )

        self.logger.info(
            f"Generated templates from origin showers in {origin_shower_repo} and saved to {template_repo}."
        )

        # make a configuration file to record all the configurations
        with open(os.path.join(path_to_templates, "config.txt"), "w") as config_file:
            config_file.write(config_txt)

    def load_templates(
        self: Self,
        template_repo: str,
        xmax_range: Union[tuple, None] = None,
        zenith_range: Union[tuple, None] = None,
        randomize: bool = False,
    ) -> None:
        """
        Load the templates for synthesis. This will load the necessary templates which will be used for the run.

        This will automatically load the templates if they are available, otherwise new templates will be generated from origin showers.

        The origin shower library will be used by default. If you want to perform synthesis with a single origin shower (e.g. for testing purposes), use the `synthesize_with_single_origin` function instead.

        If you only want to load a subset of xmax / zenith values for the synthesis, you can do so by specifying the ranges. If no ranges are provided, all available templates will be loaded.

        Parameters:
        ----------
        template_repo : str
            The path to the repository containing the origin shower library (templates).
        xmax_range : Union[tuple, None], default=None
            The range of xmax values to load templates for. If None, all xmax values will be loaded.
        zenith_range : Union[tuple, None], default=None
            The range of zenith angles to load templates for. If None, all zenith angles will be loaded.
        randomize : bool, default=False
            Whether to randomly select one template per configuration. Default is False.

        Returns:
        --------
        None
        """
        # first check if the template repo exists
        if not os.path.exists(template_repo):
            raise FileNotFoundError(
                f"Template repository {template_repo} does not exist."
            )

        # get the path to all the templates
        path_to_templates = os.path.join(template_repo, self.site, self.backend)

        # get all configurations of xmax and zenith
        config_data = np.genfromtxt(
            os.path.join(path_to_templates, "config.txt"), delimiter=",", skip_header=1
        )
        all_xmaxs = config_data[:, 0]
        all_zeniths = config_data[:, 1]

        if xmax_range is None or zenith_range is None:
            self.logger.warning(
                "No xmax_range or zenith_range provided. Loading all available templates. This may be memory intensive (> 10 GB)."
            )

        if xmax_range is None:
            xmax_range = (min(all_xmaxs), max(all_xmaxs))
        if zenith_range is None:
            zenith_range = (min(all_zeniths), max(all_zeniths))

        all_template_paths = self.__get_all_templates(
            path_to_templates, xmax_range, zenith_range, randomize
        )
        self.origin_xmaxs = np.zeros(len(all_template_paths))
        self.origin_zeniths = np.zeros(len(all_template_paths))

        for itemp, template_path in enumerate(all_template_paths):
            synthesis = self.smiet.TemplateSynthesis(freq_ar=self.freq_ar)
            synthesis.load_template(template_path)

            self.syntheses.append(synthesis)
            self.origin_xmaxs[itemp] = synthesis.template_information["xmax"]
            self.origin_zeniths[itemp] = np.rad2deg(
                synthesis.template_information["zenith"]
            )

        # sort in increasing xmax
        sort_indices = np.argsort(self.origin_xmaxs)
        self.origin_xmaxs = self.origin_xmaxs[sort_indices]
        self.origin_zeniths = self.origin_zeniths[sort_indices]
        self.syntheses = [self.syntheses[i] for i in sort_indices]

        # also set the delta_t here, it will be the same for all origins
        self.delta_t = self.syntheses[0].delta_t

        # give a log message
        self.logger.info(
            f"Loaded {len(self.syntheses)} templates for synthesis from {template_repo} within xmax range {xmax_range} and zenith range {zenith_range}."
        )

        self.__loaded_templates = True

    def __get_all_templates(
        self: Self,
        path_to_templates: str,
        xmax_range: tuple,
        zenith_range: tuple,
        randomize: bool = False,
    ) -> list[str]:
        """
        Get all templates names that are within the range of xmax and zenith angles.

        Parameters:
        ----------
        path_to_templates : str
            The path to the templates.
        xmax_range : tuple
            The range of xmax values to filter templates.
        zenith_range : tuple
            The range of zenith angles to filter templates.
        randomize : bool, default=False
            Whether to randomly select one template per configuration. Default is False.
        """
        all_templates = []

        rng = np.random.default_rng()

        for xmax_zenith in os.listdir(path_to_templates):
            if not os.path.isdir(os.path.join(path_to_templates, xmax_zenith)):
                continue
            config_path = os.path.join(path_to_templates, xmax_zenith)

            # get xmax and zenith from naming structure
            xmax_str, zenith_str = xmax_zenith.split("_")
            xmax = float(xmax_str.replace("xmax", ""))
            zenith = float(zenith_str.replace("zen", ""))

            n_sims_per_config = len(os.listdir(config_path))

            if (
                xmax >= xmax_range[0]
                and xmax <= xmax_range[1]
                and zenith >= zenith_range[0]
                and zenith <= zenith_range[1]
            ):
                # take a random number, since there are some number of simulations per directory
                rand_sim = (
                    rng.integers(low=1, high=n_sims_per_config + 1) if randomize else 1
                )
                template_path = os.path.join(
                    path_to_templates, xmax_zenith, f"template_sim{rand_sim}.hdf5"
                )

                all_templates.append(template_path)

        # make sure that its not empty
        if len(all_templates) == 0:
            raise ValueError(
                f"No templates found in {template_path} for the given xmax range {xmax_range} and zenith range {zenith_range}."
            )

        return all_templates

    def load_coreas_showers(
        self: Self,
        coreas_shower_paths: list[str],
        gdas_atmosphere_paths: Union[list[str], None] = None,
    ):
        """
        Load target showers in the form of existing coreas simulations.

        Parameters:
        -----------
        coreas_shower_paths : list[str]
            A list of paths to each coreas simulation in HDF5 format. Each path corresponds to a single target shower.
        gdas_atmosphere_paths : Union[list[str], None], default=None
            The path to the GDAS atmosphere for each simulation. Default is None, where the model set globally is used. If provided, the entries and length of this list must match the length of coreas_shower_paths.
        """
        self.target_xmaxs = np.zeros(len(coreas_shower_paths))
        self.target_zeniths = np.zeros(len(coreas_shower_paths))

        for ishower, coreas_shower_path in enumerate(coreas_shower_paths):
            # make sure paths exist
            if not os.path.exists(coreas_shower_path):
                raise FileNotFoundError(
                    f"Coreas shower path {coreas_shower_path} does not exist."
                )

            gdas_atmosphere_path = (
                gdas_atmosphere_paths[ishower]
                if gdas_atmosphere_paths is not None
                else None
            )

            target_shower = self.smiet.CoreasShower(
                coreas_shower_path, gdas_file=gdas_atmosphere_path
            )

            self.target_xmaxs[ishower] = target_shower.xmax
            self.target_zeniths[ishower] = np.rad2deg(target_shower.zenith)
            self.targets.append(target_shower)

        # raise error if target xmaxs or zeniths are outside the range of origin showers
        if not self.__check_xmax_zenith_range():
            raise ValueError(
                "Target shower xmax or zenith values are outside the range of origin shower values. Load templates with a wider xmax / zenith range."
            )

        self.logger.info(
            f"Loaded {len(self.targets)} target showers from Coreas simulations."
        )
        self.__loaded_targets = True

    def __check_xmax_zenith_range(self: Self) -> bool:
        """
        Check whether the xmax and zenith of the target showers are within the range of the origin showers.
        """
        within_xmax = min(self.target_xmaxs) > min(self.origin_xmaxs) and max(
            self.target_xmaxs
        ) < max(self.origin_xmaxs)
        within_zenith = min(self.target_zeniths) > min(self.origin_zeniths) and max(
            self.target_zeniths
        ) < max(self.origin_zeniths)

        return within_xmax and within_zenith

    def generate_target_showers(
        self: Self,
        target_grammage: np.ndarray,
        target_longs: list[np.ndarray],
        target_zeniths: np.ndarray,
        target_azimuths: Union[np.ndarray, None] = None,
        target_cores: Union[np.ndarray, None] = None,
    ) -> None:
        """
        Generate and set target showers based on user-provided longitudinal profiles and geometries.

        Parameters:
        -----------
        target_grammage : np.ndarray
            The grammage slices corresponding to the provided longitudinal profiles. Should be a numpy array of shape (N,) where N is the number of grammage slices.
        target_longs : list[np.ndarray]
            A list of longitudinal profiles for each target shower. Each profile should be a numpy array of shape (N,) where N is the number of grammage slices. The number of grammage slices need not match those in the templates, as interpolation will be performed internally.
        target_zeniths : np.ndarray
            The zenith angles for each target shower. Should be a numpy array of shape (M,) where M is the number of target showers.
        target_azimuths : Union[np.ndarray, None], default=None
            The azimuth angles for each target shower. Should be a numpy array of shape (M,) where M is the number of target showers. If None, azimuths will be set to 0 degrees by default.
        target_cores : Union[np.ndarray, None], default=None
            The core positions for each target shower. Should be a numpy array of shape (M, 2) where M is the number of target showers and each entry contains the (x, y) core position in meters. If None, cores will be set to (0, 0) by default.
        """
        target_azimuths = (
            np.zeros(len(target_longs)) if target_azimuths is None else target_azimuths
        )
        target_cores = (
            np.zeros((len(target_longs), 3)) if target_cores is None else target_cores
        )

        if self.backend == "jax":
            target_zeniths = jnp.array(target_zeniths)
            target_azimuths = jnp.array(target_azimuths)
            target_cores = jnp.array(target_cores)
            target_grammage = jnp.array(target_grammage)

        self.target_xmaxs = np.zeros(len(target_longs))
        for ishower, long in enumerate(target_longs):
            # should try to fit GH to get xmax and nmax
            # since argmax and max are limited by the
            # grid size.
            # for now its fine, we will adjust this later
            # in any case, for the np version the GH parameters will be fit
            xmax = target_grammage[long.argmax()]
            nmax = long.max()

            target_shower = self.smiet.Shower()

            target_shower.set_parameters(
                grammages=target_grammage,
                long_profile=long,
                xmax=xmax,
                nmax=nmax,
                zenith=target_zeniths[ishower],
                azimuth=target_azimuths[ishower],
                core=target_cores[ishower],
                magnetic_field=self.magnet,
            )

            self.target_xmaxs[ishower] = target_shower.xmax
            self.targets.append(target_shower)

        self.target_zeniths = np.rad2deg(target_zeniths)

        # raise error if target xmaxs or zeniths are outside the range of origin showers
        if not self.__check_xmax_zenith_range():
            raise ValueError(
                "Target shower xmax or zenith values are outside the range of origin shower values. Load templates with a wider xmax / zenith range."
            )

        self.logger.info(
            f"Generated {len(self.targets)} target showers from user-provided longitudinal profiles."
        )
        self.__loaded_targets = True

    def synthesize_showers(
        self: Self,
        interpolation: bool = True,
        slices: bool = False,
    ) -> None:
        """
        Run the synthesis for all loaded target showers.

        This will perform the interpolation synthesis method by default, where the combination of two synthesis, weighted by some function, will be used. It is most adviced to use this method, so only disable the interpolation if you need to.

        Parameters:
        -----------
        interpolation : bool, default=True
            Whether to use interpolation synthesis method. If False, the nearest neighbor synthesis object will be used.
        slices : bool, default=False
            Whether to synthesise traces for all slices. If False, only synthesises at ground level.
        """
        # first make sure all flags are checked
        if not self.__loaded_templates:
            raise RuntimeError(
                "Templates have not been loaded. Please load templates before running synthesis."
            )
        if not self.__loaded_targets:
            raise RuntimeError(
                "Target showers have not been loaded. Please load target showers before running synthesis."
            )
        if slices:
            self.logger.warning(
                "Synthesis with slices=True is enabled. Setting interpolation=False since it does not work with interpolated synthesis."
            )
            interpolation = False

        for ishower, target_shower in enumerate(self.targets):
            # find the closest origin shower, the smaller one
            # for zenith, synthesis only works if theta_origin > theta_target
            # otherwise synthesis will artificially generate showers
            diff_xmaxs = self.origin_xmaxs - target_shower.xmax
            diff_zenith = self.origin_zeniths - target_shower.zenith
            lower_indices = np.where(np.logical_and(diff_xmaxs <= 0, diff_zenith > 0))[
                0
            ]
            upper_indices = np.where(np.logical_and(diff_xmaxs > 0, diff_zenith > 0))[0]

            if len(lower_indices) == 0 or len(upper_indices) == 0:
                raise ValueError(
                    f"Target shower xmax {target_shower.xmax} is outside the range of origin shower xmax values."
                )

            lower_index = lower_indices[-1]
            upper_index = upper_indices[0]

            lower_synth = self.syntheses[lower_index]
            upper_synth = self.syntheses[upper_index]

            # here we get the synthesised traces on a star shape
            # from a vxvxB synthesis
            target_shower.transform_profile_to_origin(lower_synth.grammages)
            synth_traces_geo_vvB, synth_traces_ce_vvB = (
                lower_synth.map_template_to_slices(target_shower)
                if slices
                else lower_synth.map_template(target_shower)
            )
            synth_traces_geo, synth_traces_ce, ant_pos_starshape, ant_labels = (
                self.smiet.utilities.reconstruct_starshape_from_vxvxB(
                    synth_traces_geo_vvB,
                    synth_traces_ce_vvB,
                    lower_synth.antenna_information["position_showerplane"][:, 1],
                )
            )

            synth_traces = np.array([synth_traces_geo, synth_traces_ce])

            if interpolation:
                target_shower.transform_profile_to_origin(upper_synth.grammages)
                synth_traces_geo_upper_vvB, synth_traces_ce_upper_vvB = (
                    upper_synth.map_template(target_shower)
                )
                synth_traces_geo_upper, synth_traces_ce_upper, _, _ = (
                    self.smiet.utilities.reconstruct_starshape_from_vxvxB(
                        synth_traces_geo_upper_vvB,
                        synth_traces_ce_upper_vvB,
                        upper_synth.antenna_information["position_showerplane"][:, 1],
                    )
                )

                synth_traces_upper = np.array(
                    [synth_traces_geo_upper, synth_traces_ce_upper]
                )

                # calculate weight factor for interpolation
                # TODO: vectorize this!
                w_xmax = (target_shower.xmax - self.origin_xmaxs[lower_index]) / (
                    self.origin_xmaxs[upper_index] - self.origin_xmaxs[lower_index]
                )

                synth_traces *= 1 - w_xmax
                synth_traces += w_xmax * synth_traces_upper

            if not slices:
                synth_traces = synth_traces[..., np.newaxis]  # add slice axis

            self.synth_traces.append(synth_traces)
            self.synth_times.append(
                np.repeat(lower_synth.get_time_axis(), repeats=8, axis=0)
            )  # 8 is number of antennas in star shape, hardcoded
            self.antenna_positions.append(ant_pos_starshape)

        # antenna labels only need to be assigned once
        self.antenna_labels = ant_labels

        self.logger.info(f"Synthesized {len(self.synth_traces)} target showers.")

        self.__synthesized_showers = True

    def __find_target_idx(self: Self, xmax: float, zenith: float) -> int:
        """
        Find the index corresponding to the target shower, given a xmax and zenith value

        Parameters:
        -----------
        xmax : float
            The xmax of the target shower to find.
        zenith : float
            The zenith angle of the target shower to find.
        """
        # find the target shower
        target_index = None
        for ishower, target_shower in enumerate(self.targets):
            if np.isclose(target_shower.xmax, xmax) and np.isclose(
                np.rad2deg(target_shower.zenith), zenith
            ):
                target_index = ishower
                break

        if target_index is None:
            raise ValueError(
                f"Target shower with xmax {xmax} and zenith {zenith} not found."
            )
        return target_index
    
    def __get_sample_axis(
        self: Self,
        ant_name: Union[float, None] = None,
        slice_grammage: Union[float, None] = None,
    ):
        """Get the sample axis based on the provided antenna name and slice grammage. Could be implemented more efficiently!"""
        if ant_name is not None:
            return 1
        elif slice_grammage is not None:
            return 2
        elif (ant_name is not None) and (slice_grammage is None):
            return 1
        elif (ant_name is None) and (slice_grammage is not None):
            return 1
        elif (ant_name is None) and (slice_grammage is None):
            return 2
        else:
            return 2
        

    def get_traces(
        self: Self,
        xmax: float,
        zenith: float,
        ant_name: Union[str, None] = None,
        geoce: bool = True,
        slice_grammage: Union[float, None] = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get the time traces for a specific target shower.

        Parameters:
        -----------
        xmax : float
            The xmax of the target shower to retrieve traces for.
        zenith : float
            The zenith angle of the target shower to retrieve traces for.
        ant_name : Union[str, None], default=None
            The name of the antenna to retrieve traces for. If None, all antennas will be returned.

            The nomenclature for antenna names is given as 'ant_r{r}_phi{phi}`, where `phi` is the azimuthal angle in degrees and `r` is the radial distance in meters.

            The avaliable antenna names can be found in the `antenna_labels` attribute after synthesis.
        geoce : bool, default=True
            Whether to return traces in the geomagnetic + charge-excess emissions. If False, returns in the shower plane coordinate system.
        slice_grammage : Union[float, None], default=None
            The grammage slice to retrieve traces for. If None, retrieves trace at the ground. If traces are only synthesised at the ground, the argument will default to None.

        Returns:
        --------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            A tuple containing:
            - traces: np.ndarray of shape (N_pol, N_antennas, N_timebins, n_slices)
                The traces either in geo/ce components (if geoce is True) or in vxB/vxvxB components (if geoce is False).
            - times: np.ndarray of shape (N_antennas, N_timebins, n_slices)
                The time axes corresponding to each antenna trace.
            - antenna_positions: np.ndarray of shape (N_antennas, 3)
                The positions of each antenna in meters in the shower plane.

            If ant_name is not None, N_antennas will be 1 and the corresponding antenna trace & time will be returned.

        """
        if not self.__synthesized_showers:
            raise RuntimeError(
                "Showers have not been synthesized yet. Please run synthesis before retrieving traces."
            )

        target_index = self.__find_target_idx(xmax, zenith)

        synth_traces = self.synth_traces[target_index]
        trace_times = self.synth_times[target_index]
        ant_positions = self.antenna_positions[target_index]

        if not geoce:
            # convert to vxB / vxvxB
            synth_traces = self.smiet.utilities.geo_ce_to_e(
                synth_traces[0],
                synth_traces[1],
                ant_positions[:, 0],
                ant_positions[:, 1],
            )

            # match the shape of {Npol, Nant, Ntime}
            synth_traces = np.swapaxes(synth_traces.T, 3, 1)

        if ant_name is not None:
            if ant_name not in self.antenna_labels:
                raise ValueError(
                    f"Antenna name {ant_name} not found in available antenna labels."
                )
            ant_index = self.antenna_labels.index(ant_name)

            synth_traces = np.take(synth_traces, indices=ant_index, axis=1)
            trace_times = np.take(trace_times, indices=ant_index, axis=0)
            ant_positions = np.take(ant_positions, indices=ant_index, axis=0)

        # force slice grammage to None if only synthesised at ground
        if synth_traces.shape[-1] == 1:
            self.logger.warning(
                "Traces were only synthesised at ground level. Ignoring provided slice_grammage and returning ground level traces."
            )
            slice_grammage = None

        if slice_grammage is not None:
            # find the closest slice
            grammages = self.targets[target_index].grammages
            slice_index = np.argmin(np.abs(grammages - slice_grammage))
            synth_traces = np.take(synth_traces, indices=slice_index, axis=-1)
        else:
            synth_traces = np.sum(synth_traces, axis=-1)

        return synth_traces, trace_times, ant_positions

    def get_spectrum(
        self: Self,
        xmax: float,
        zenith: float,
        ant_name: Union[str, None] = None,
        geoce: bool = True,
        slice_grammage: Union[float, None] = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Get the frequency spectrum for a given synthesised shower and antenna name (optional).

        Parameters:
        -----------
        xmax : float
            The xmax of the target shower to retrieve spectrum for.
        zenith : float
            The zenith angle of the target shower to retrieve spectrum for.
        ant_name : Union[str, None], default=None
            The name of the antenna to retrieve spectrum for. If None, all antennas will be returned.
        geoce : bool, default=True
            Whether to return traces in the geomagnetic + charge-excess emissions. If False, returns in the shower plane coordinate system.
        slice_grammage : Union[float, None], default=None
            The grammage slice to retrieve traces for. If None, retrieves traces at all slices. If traces are only synthesised at the ground, then that will be returned instead.

        Returns:
        --------
        tuple[np.ndarray, np.ndarray]
            A tuple containing:
            - spectrum: np.ndarray of shape (N_pol, N_antennas, N_freqbins, Nslices)
                The frequency spectrum either in geo/ce components (if geoce is True) or in vxB/vxvxB components (if geoce is False).
            - freq_grid: np.ndarray of shape (N_freqbins,)
                The frequency axis corresponding to the spectrum.

        If ant_name is not None, N_antennas will be 1 and the corresponding antenna spectrum will be returned.
        """
        # basically call the get_traces function to get the relevant traces
        synth_traces, trace_times, _ = self.get_traces(
            xmax, zenith, ant_name=ant_name, geoce=geoce, slice_grammage=slice_grammage
        )
        sample_idx = self.__get_sample_axis(ant_name, slice_grammage)
        # then perform an FFT to get the frequency spectrum
        synth_spect = np.fft.rfft(synth_traces, axis=sample_idx, norm="ortho", n=synth_traces.shape[sample_idx])
        synth_freq_grid = np.fft.rfftfreq(synth_traces.shape[sample_idx], d=self.delta_t)

        return synth_spect, synth_freq_grid

    def get_fluence(
        self: Self, xmax: float, zenith: float, geoce: bool = True, slice_grammage: Union[float, None] = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Get the fluence for a given synthesised shower.

        Parameters:
        ----------
        xmax : float
            The xmax of the target shower to retrieve fluence for.
        zenith : float
            The zenith angle of the target shower to retrieve fluence for.
        geoce : bool, default=True
            Whether to return traces in the geomagnetic + charge-excess emissions. If False, returns in the shower plane coordinate system.
        slice_grammage : Union[float, None], default=None
            The grammage slice to retrieve traces for. If None, retrieves traces at all slices. If traces are only synthesised at the ground, then that will be returned instead.
        """
        # again call the get_traces function to get the relevant traces
        # but here call without antenna name to get all antennas
        synth_traces, _, ant_positions = self.get_traces(
            xmax, zenith, ant_name=None, geoce=geoce, slice_grammage=slice_grammage
        )
        sample_idx = self.__get_sample_axis(ant_name=None, slice_grammage=slice_grammage)
        # then perform a fluence calculation
        conversion_factor_integrated_signal = (
            2.65441729e-3 * 6.24150934e18
        )  # to convert to eV/m^2 from (V/m)^2 s
        synth_fluence = (
            np.sum(synth_traces**2, axis=(0, sample_idx))
            * (self.delta_t / units.s)
            * conversion_factor_integrated_signal
        )
        return synth_fluence, ant_positions

    def get_truths(
        self: Self,
        xmax: float,
        zenith: float,
        mode: str = "fluence",
        geoce: bool = True,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Return the truth values for a given synthesised shower.

        Note that this is only available if the target showers were loaded from Coreas simulations. Otherwise, an error will be raised.

        Parameters:
        ----------
        xmax : float
            The xmax of the target shower to retrieve truths for.
        zenith : float
            The zenith angle of the target shower to retrieve truths for.
        mode : str, default="fluence"
            The mode of truth values to retrieve. Options are "fluence", "spectrum", or "traces".
        geoce : bool, default=True
            Whether to return traces in the geomagnetic + charge-excess emissions. If False, returns in the shower plane coordinate system.
        """
        # first check if target showers are coreas showers
        if not isinstance(self.targets[0], self.smiet.CoreasShower):
            raise RuntimeError(
                "Target showers are not Coreas showers. Truth values are only available for Coreas showers."
            )
        target_index = self.__find_target_idx(xmax, zenith)
        ant_names = self.targets[target_index].antenna_names
        ant_pos_showerplane = self.targets[
            target_index
        ].get_antenna_position_showerplane(ant_names)

        if self.backend == "numpy":
            if geoce:
                traces_truth_geo, traces_truth_ce, traces_times = self.targets[
                    target_index
                ].get_traces_geoce(
                    ant_names=ant_names,
                    return_times=True,
                )
                traces_truth = np.array([traces_truth_geo, traces_truth_ce])
            else:
                traces_truth_vB, traces_truth_vvB, traces_times = self.targets[
                    target_index
                ].get_traces_vB_vvB(
                    ant_names=ant_names,
                    return_times=True,
                )
                traces_truth = np.array([traces_truth_vB, traces_truth_vvB])
        else:
            raise NotImplementedError("will implement it later.")

        if mode == "traces":
            return traces_truth, traces_times, ant_pos_showerplane
        elif mode == "spectrum":
            # then perform an FFT to get the frequency spectrum
            spect_truth = np.fft.rfft(traces_truth, axis=-1, norm="ortho")
            freq_grid = np.fft.rfftfreq(traces_times.shape[-1], d=self.targets[target_index].delta_t)
            return spect_truth, freq_grid, ant_pos_showerplane
        elif mode == "fluence":
            # then perform a fluence calculation
            conversion_factor_integrated_signal = 2.65441729e-3 * 6.24150934e18
            fluence_truth = (
                np.sum(traces_truth**2, axis=(0, -1))
                * (self.targets[target_index].delta_t / units.s)
                * conversion_factor_integrated_signal
            )
            return fluence_truth, ant_pos_showerplane
        else:
            raise ValueError(
                f"Mode {mode} not recognized. Options are 'fluence', 'spectrum', or 'traces'."
            )

    def synthesize_with_single_origin(
        self: Self,
        origin_shower_path: str,
        target_long: np.ndarray,
        target_grammage: np.ndarray,
        target_zenith: float,
        target_azimuth: float = 0.0,
        target_core: np.ndarray = np.array([0.0, 0.0, 0.0]),
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Perform synthesis with a single origin shower.

        This function is mostly for testing purposes, but is available if users want to do this.

        Parameters:
        ----------
        origin_shower_path : str
            The path to the origin shower to use for synthesis.
        target_long : np.ndarray
            The longitudinal profile of the target shower.
        target_grammage : np.ndarray
            The grammage slices corresponding to the provided longitudinal profile.
        target_zenith : float
            The zenith angle of the target shower.
        target_azimuth : float, default=0.0
            The azimuth angle of the target shower. Default is 0.0 degrees.
        target_core : np.ndarray, default=np.array([0.0, 0.0, 0.0])
            The core position of the target shower in meters. Default is (0.0, 0.0, 0.0).

        Returns:
        --------
        np.ndarray
            The synthesised time traces for the target shower and the time axes.
        """
        # load an origin shower
        if not os.path.exists(origin_shower_path):
            raise FileNotFoundError(
                f"Origin shower path {origin_shower_path} does not exist."
            )
        origin_shower = self.smiet.SlicedShower(origin_shower_path)

        # make a template
        synthesis = self.smiet.TemplateSynthesis(freq_ar=self.freq_ar)
        synthesis.make_template(origin_shower)

        # generate a target shower
        target_shower = self.smiet.Shower()
        target_shower.set_parameters(
            grammages=target_grammage,
            long_profile=target_long,
            xmax=target_grammage[target_long.argmax()],
            nmax=target_long.max(),
            zenith=np.deg2rad(target_zenith),
            azimuth=np.deg2rad(target_azimuth),
            core=target_core,
            magnetic_field=self.magnet,
        )

        target_shower.transform_profile_to_origin(origin_shower.grammages)

        # map the template
        synth_traces_geo, synth_traces_ce = synthesis.map_template(target_shower)

        # return the synthesised traces on the vxvxB arm (or starshape?)
        return synth_traces_geo, synth_traces_ce, synthesis.get_time_axis()

    def target_properties(self: Self) -> dict:
        """
        Return the properties of each target shower
        """
        return {
            "xmaxs": self.target_xmaxs,
            "zeniths": self.target_zeniths,
            "long": [self.targets[i].long_profile for i in range(len(self.targets))],
            "grammage": [self.targets[i].grammages for i in range(len(self.targets))],
        }

    def origin_properties(self: Self) -> dict:
        """
        Return the properties of each origin shower
        """
        return {
            "xmaxs": self.origin_xmaxs,
            "zeniths": self.origin_zeniths,
            "long": [
                self.syntheses[i].template_information["long_profile"]
                for i in range(len(self.syntheses))
            ],
            "grammage": [
                self.syntheses[i].grammages for i in range(len(self.syntheses))
            ],
        }



import logging
from typing_extensions import Self, Tuple, Union

import numpy as np
import h5py
import jax

jax.config.update("jax_enable_x64", True)

from smiet import units
from jax_radio_tools import (
    e_to_geo_ce,
    conversion_fieldstrength_cgs_to_SI,
)
from .coreas_shower import CoreasShower

logger = logging.getLogger("smiet.jax.io")


class SlicedShower(CoreasShower):
    """
    Class to read in showers from each slice from CoREAS simulations, inherited from CoreasShower.

    Parameters
    ----------
    file_path : str
        The filepath to the simulation to read.
    slicing_grammage : int, default=5
        The width between atmospheric slices in g/cm^2
    bandpass : list, default = [30 * units.MHz, 500 * units.MHz]
        The frequency bandpass in MHz to apply to the traces. Defaults to 30-500 MHz,
        which is the bandwidth in which SMIET is applicable in.
    gdas_file : str, optional
        If provided, the atmosphere will be created using this GDAS file. This will take
        precedence over the atmosphere model specified in the simulation settings.
    """

    def __init__(
        self: Self,
        file_path: str,
        slicing_grammage: int = 5,  # not used since its read from the file itself
        bandpass: list = [30 * units.MHz, 500 * units.MHz],
        gdas_file : str = "",
    ) -> None:
        
        self.logger = logging.getLogger("smiet.jax.io.SlicedShower")
        self.__slice_gram = slicing_grammage  # g/cm2, TODO: change how to read this since its fixed from filepath  # Not used
        super().__init__(file_path, bandpass=bandpass, gdas_file=gdas_file)



    def _load_antenna_and_traces(self : Self, file : h5py.File) -> None:
        """Read in the antenna information from the HDF5 file."""
        # observer properties
        self.antenna_array["name"] = list(
            sorted(
                set([key.split("x")[0] for key in file["CoREAS"]["observers"].keys()])
            )
        )

        # antenna positions at the ground
        self.antenna_array["position"] = np.array(
            [
                file["CoREAS"]["observers"][f"{ant_name}x{self.__slice_gram}"].attrs[
                    "position"
                ]
                * units.cm
                for ant_name in self.antenna_array["name"]
            ]
        )  # antennas x 3

        self.antenna_array["position"] = self.antenna_array["position"][:, [1, 0, 2]]
        self.antenna_array["position"][:,0] *= -1

        # distance from the shower core
        self.antenna_array["dis_to_core"] = np.linalg.norm(
            (self.get_antenna_position_showerplane()[:, :2]), axis=-1
        )

        # the observer data for each slice, containing both timing and trace information
        observer_slices = np.array(
            [
                [
                    file["CoREAS"]["observers"][f"{ant_name}x{int(gram):d}"][:]
                    for ant_name in self.antenna_array["name"]
                ]
                for gram in self.grammages
            ]
        )  # slices x antennas x samples x 4

        # map it to the correct shape of
        # 4 x ANT x SAMPLES x SLICE as according to numpy version
        observer_slices = np.moveaxis(observer_slices, (0, 1, 2, 3), (3, 1, 2, 0))

        # divide into trace times & actual E-field slices
        # we also do not care about the time for each slice, since they arrive at the same time
        self.trace_times = (
            observer_slices[0, ..., -1] * units.s
        )  # shape of (Nant, Nsamples), not needed but keep incase
        self.traces = observer_slices[
            1:, ...
        ]  # shape of (Npol, Nant, Nsamples, Nslices)

        self.trace_length = self.traces.shape[2]

        self.traces = self.traces[[1, 0, 2], ...] * conversion_fieldstrength_cgs_to_SI  # convert traces from cgs to SI manually (due to electric charge shenanigans)
        self.traces[0,...] *= -1

    def get_coreas_settings(self: Self) -> dict:
        """Get specific configurations from the CoREAS simulation that is useful for the synthesis.

        Returns
        -------
        coreas_settings : dict
            A dictionary containing important information about the configuration of the CoREAS simulation
        """
        file = h5py.File(self.__file)

        time_resolution = float(file["CoREAS"].attrs["TimeResolution"]) * units.s

        return {"time_resolution": time_resolution}
    
    def get_traces_geoce(self: Self) -> np.ndarray:
        """
        Get the traces from traces at the ground to GEO/CE components.

        Returns
        -------
        traces_geo_ce : np.ndarray
            The traces in the shower plane, transformed to GEO and CE components.
            Shape is (2, Nant, Nsamples), where Nant is the number of antennas and Nsamples is the number of samples.

            When called with SlicedShower, the shape is (2, Nant, Nsamples, Nslices).
        """
        traces_vvB = self.get_traces_vB_vvB()
        ant_position_vvB = self.get_antenna_position_showerplane()
        # now convert trace to respective components
        traces_geo, traces_ce = e_to_geo_ce(
            traces_vvB.T, ant_position_vvB[None, None, :, 0], ant_position_vvB[None, None, :, 1]
        )
        return np.array([traces_geo.T, traces_ce.T])



import os, shutil
import numpy as np
import h5py
from typing import Union
import json

import radiotools.helper as hp
from radiotools.atmosphere.models import Atmosphere

from smiet import units
from smiet.corsika.generate_file_contents import (
    generate_list_file,
    generate_reas_file,
    generate_inp_file,
    generate_list_file_cherenkov,
    generate_list_file_from_positions,
)

import logging

logger = logging.getLogger("smiet.corsika.write_simulation")


__all__ = [
    "generate_simulation",
    "generate_simulation_from_hdf5",
    "write_simulation_to_file",
    "sample_primary_energy",
    "sample_zenith_angle",
    "sample_azimuth_angle",
]


CORSIKA_PARTICLE_CODES = {
    "proton": 14,
    "helium": 402,
    "carbon": 1206,
    "silicon": 2814,
    "iron": 5626,
}

# some default values for the first height parameterisation
xmax_min = 400.0  # minimum valid range for xmax in g/cm^2
xmax_max = 1300.0  # maximum valid range for xmax in g/cm^2
h_max = 112829.2 * units.m  # maximum height of the atmosphere model

# Functions to generate all files for a single simulation


def generate_simulation(
    sim_nr,
    sim_primary,
    sim_cores,
    sim_zenith=None,
    sim_azimuth=None,
    sim_energy=None,
    sim_xmax=None,
    slice_gram=5.0,
    magnetic_field="lofar",
    thinning=1e-7,
    atmosphere=17,
    core=None,
    radii=None,
    nr_of_arms=1,
    cherenkov=False,
    rotate : bool = False,
):
    """
    Prepare the contents for the INP, LIST and REAS files for a sliced MPI simulation, with a star shape pattern for the antenna layout.

    One can choose to project the antennas along the shower axis or along the viewing angle by
    setting the `cherenkov` option. If this is set to True, the `radii` parameter will be interpreted
    as a list of viewing angles expressed as multiples of the slice Cherenkov angle. Otherwise,
    the `radii` parameter should be the list of radii to use in the star shape
    (use this option to generate origin showers).

    The magnetic field vector can be taken from the `radiotools.helper` module, by passing a string
    as the `magnetic_field` parameter. This variable is passed on to the
    `radiotools.helper.get_magnetic_field_vector` function. But to be consistent with the
    CORSIKA definition where the magnetic field has no East-West component, the vector x-component
    is set to zero. This ensures the simulated antennas are indeed on the (vx)vxB axis.

    The primary particle can be specified as a string, which will be converted to the corresponding
    CORSIKA particle code. As of June 2025, only proton, helium, carbon, silicon and iron primaries
    are supported (this is limited by the CORSIKA_PARTICLE_CODES dictionary).

    Parameters
    ----------
    sim_nr : int
        The number of the simulation.
    sim_primary : {'proton', 'helium', 'carbon', 'silicon', 'iron'}
        The primary particle to inject
    sim_cores : int
        The number of cores the simulation will be run on
    sim_zenith : float, optional
        The zenith angle in degrees. If not provided, the angle will be drawn from `sample_zenith_angle`.
    sim_azimuth : float, optional
        The azimuth angle in degrees. If not provided, the angle will be drawn from `sample_azimuth_angle`.
    sim_energy : float, optional
        The energy of the simulation in GeV. If not provided, the energy will be drawn from `sample_primary_energy`.
    sim_xmax : float, optional
        The depth of shower maximum in g/cm^2. If provided, this value (along with sim_energy and sim_zenith)
        will be used to compute the first interaction height of the shower, using the parameterisation from CONEX.
        This value will be passed in the FIXHEI parameter in the CORSIKA input file.
    core : list of float, default=[0, 0, 0] * units.m
        The core to use in the simulation
    slice_gram : float, default=5.0
        The thickness of the atmospheric slices.
        NB: DO NOT put in internal units. We default the units to be g/cm^2.
    radii : list of float, default=None
        The radii of the starshape, passed to `smiet.corsika.generate_list_file`
    nr_of_arms: int, default=1
        The number of arms in the star shape pattern. This is passed to `smiet.corsika.generate_list_file`.
    cherenkov : bool, default=False
        If True, interpret `radii` as multiples of slice Cherenkov radius, and calculate antenna radii per slice
    magnetic_field : np.ndarray or str, default='lofar'
        If a string, the name of the site from which to take the magnetic field vector (using radiotools.helper).
        Otherwise, this is interpreted as the magnetic field vector (in internal units).
    thinning : float, default=1e-7
        The thinning level to use
    atmosphere : int or str, default=17
        The CORSIKA atmosphere identifier. If a string, the path to the GDAS atmosphere file.
    rotate : bool, default=False
        Whether to rotate the antenna positions in the LIST file by 22.5 degrees so that the do not
        align with the vxB axis, which we know fails with template synthesis.
    """
    assert sim_primary in CORSIKA_PARTICLE_CODES, (
        f"Primary {sim_primary} not supported!"
    )

    if core is None:
        core = np.array([0.0, 0.0, 0.0]) * units.m

    sim_zenith = sample_zenith_angle()[0] if sim_zenith is None else sim_zenith
    sim_azimuth = sample_azimuth_angle()[0] if sim_azimuth is None else sim_azimuth
    sim_primary_energy = (
        sample_primary_energy()[0] if sim_energy is None else sim_energy
    )
    logger.info(f"Zenith angle : {sim_zenith / units.deg:.2f} deg")
    logger.info(f"Azimuth angle : {sim_azimuth / units.deg:.2f} deg")
    logger.info(f"Primary energy : {sim_primary_energy / units.GeV:.2e} GeV")

    atm = Atmosphere(atmosphere, curved=False)
    grammage_ground = atm.get_distance_xmax(
        sim_zenith / units.rad, 0.0, observation_level=core[2] / units.m
    )
    grammage_slices = np.arange(
        0.0, (grammage_ground // slice_gram + 1) * slice_gram + 1, slice_gram
    )

    if type(magnetic_field) is str:
        magnetic_field_vector = (
            hp.get_magnetic_field_vector(magnetic_field)
            * np.array([0, 1, 1])
            * units.gauss
        )
    else:
        magnetic_field_vector = np.asarray(magnetic_field)

    # only compute the first interaction height if xmax is not None
    if sim_xmax is not None:
        # make sure that the magnetic field object is string and a site
        if not isinstance(magnetic_field, str):
            raise ValueError(
                "To compute the first interaction height, the magnetic_field parameter needs to be a string indicating the site!"
            )
        sim_first_height = get_first_height(
            mass_type=sim_primary,
            xmax=sim_xmax,
            log10E=np.log10(sim_primary_energy / units.eV),
            zenith=sim_zenith / units.rad,
            atm=atm,
            obs_lvl=core[2],
            site=magnetic_field
        )
        logger.info(
            f"Setting first interaction height to {sim_first_height / units.m:.2f} m"
        )
    else:
        sim_first_height = None

    inp_file = generate_inp_file(
        sim_nr,
        sim_zenith,
        sim_azimuth,
        sim_primary_energy,
        obs_lev=core[2],
        primary=CORSIKA_PARTICLE_CODES[sim_primary],
        n_cores=sim_cores,
        thin=thinning,
        atm=atmosphere,
        magnet=magnetic_field_vector,
        slicing=slice_gram,
        first_height=sim_first_height,
    )
    if cherenkov:
        list_file = generate_list_file_cherenkov(
            sim_zenith,
            sim_azimuth,
            grammage_slices,
            atm,
            radii,
            magnet_vector=magnetic_field_vector,
            core=core,
            number_of_arms=nr_of_arms,
        )
    else:
        list_file = generate_list_file(
            sim_zenith,
            sim_azimuth,
            grammage_slices,
            magnet_vector=magnetic_field_vector,
            core=core,
            radii=radii,
            number_of_arms=nr_of_arms,
            rotate=rotate,
        )
    reas_file = generate_reas_file(sim_nr, core=core)

    return inp_file, list_file, reas_file


def generate_simulation_from_hdf5(
    base_sim_file: str,
    sim_nr: int,
    sim_cores: int,
    sim_primary: Union[str, int, None] = None,
    sim_energy: Union[float, None] = None,
    slice_gram: float = 5.0,
    atm_file: Union[str, None] = None,
):
    """
    Generate simulation files for an origin shower from an existing coreas simulation.

    The geometry, primary particle type and energy are taken from the `base_sim_file`
    simulation, as well as the observation level, magnetic field and thinning level.
    If there is no GDAS atmosphere file provided, the atmosphere model number
    is also taken from the `base_sim_file`.

    Parameters
    ----------
    base_sim_file : str
        Path to the HDF5 simulation file on which we base the new simulation.
    sim_nr : int
        The number of the new simulation.
    sim_cores : int
        The number of cores the simulation will be run on
    sim_primary: str or int, default=None
        If provided, the primary particle type to use in the new simulation.
        If a string, the type is converted to the CORSIKA identification code using
        `CORSIKA_PARTICLE_CODES` dictionary. In this case the particle needs to be
        supported (see the documentation of :func:`generate_simulation`). Can also
        be an integer, in which case it is used as the CORSIKA particle code directly.
        If None, the primary particle type will be taken from the `base_sim_file`.
    sim_energy: float, default=None
        If provided, the primary energy to use in the new simulation. Otherwise,
        the energy will be taken from the `base_sim_file`.
    slice_gram : float, default=5.0
        The thickness of the atmospheric slices in g/cm^2.
        NB: DO NOT put in internal units. We default the units to be g/cm^2.
    atm_file : str, default=None
        The path to the GDAS atmosphere file. If None, the atmosphere model number
        from the `base_sim_file` will be used instead, and an atmosphere file will not be used.
    """
    logger.info(f"Generating simulation from existing simulation {base_sim_file}")

    coreas_h5_reader = h5py.File(base_sim_file, "r")

    # extract the relevant parameters
    # NOTE: always make sure to read from "inputs" and not "CoREAS" attributes
    # since those are not used in the new coreas version
    sim_zenith = coreas_h5_reader["inputs"].attrs["THETAP"][0] * units.deg
    # 270 shift required since it is being applied in the inp file generator
    sim_azimuth = (coreas_h5_reader["inputs"].attrs["PHIP"][0] + 270.0) * units.deg

    sim_primary_energy = (
        coreas_h5_reader["inputs"].attrs["ERANGE"][0] * units.GeV
        if sim_energy is None
        else sim_energy
    )
    if sim_primary is None:
        sim_primary = int(coreas_h5_reader["inputs"].attrs["PRMPAR"])
    elif isinstance(sim_primary, str):
        sim_primary = CORSIKA_PARTICLE_CODES[sim_primary]

    sim_obs_lev = coreas_h5_reader["inputs"].attrs["OBSLEV"] * units.cm
    sim_thinning = coreas_h5_reader["inputs"].attrs["THIN"][0]

    # read in the atmosphere file as a GDAS file if it is not None
    if atm_file is not None:
        sim_atmosphere = atm_file
    else:
        sim_atmosphere = int(coreas_h5_reader["inputs"].attrs["ATMOD"])

    # for magnetic field vector, the input file generator reads [1, -2]
    # so we need to append a zero element in the first entry, then
    # negate the last element
    sim_magnetic_field_vector = np.zeros(3)
    sim_magnetic_field_vector[1:] = (
        coreas_h5_reader["inputs"].attrs["MAGNET"] * units.microtesla
    )
    sim_magnetic_field_vector[2] *= -1

    # Generate INP file
    inp_file = generate_inp_file(
        sim_nr,
        sim_zenith,
        sim_azimuth,
        sim_primary_energy,
        obs_lev=sim_obs_lev,
        primary=sim_primary,
        n_cores=sim_cores,
        thin=sim_thinning,
        atm=sim_atmosphere,
        magnet=sim_magnetic_field_vector,
        slicing=slice_gram,
    )

    # for generating the lst file, we need to first get the grammage slices
    atm = Atmosphere(sim_atmosphere, curved=False)

    grammage_ground = atm.get_distance_xmax(
        sim_zenith / units.rad,
        0.0,
        observation_level=sim_obs_lev / units.m,
    )
    grammage_slices = np.arange(
        0.0, (grammage_ground // slice_gram + 1) * slice_gram + 1, slice_gram
    )

    # generate the LIST file
    ant_positions = np.asarray(
        [
            obs.attrs["position"]
            for obs in coreas_h5_reader["CoREAS"]["observers"].values()
        ]
    )
    ant_positions = ant_positions[:, [1, 0, 2]]
    ant_positions[:, 0] *= -1  # flip West to East to match NRR definition

    list_file = generate_list_file_from_positions(ant_positions, grammage_slices)

    # generate the REAS file
    coreas_attrs = coreas_h5_reader["CoREAS"].attrs

    core = np.array(
        [
            -coreas_attrs["CoreCoordinateWest"],
            coreas_attrs["CoreCoordinateNorth"],
            coreas_attrs["CoreCoordinateVertical"],
        ]
    )

    reas_file = generate_reas_file(
        sim_nr,
        n=coreas_attrs["GroundLevelRefractiveIndex"],
        time_res=coreas_attrs["TimeResolution"] * units.s,
        time_bound=coreas_attrs["AutomaticTimeBoundaries"] * units.s,
        core=core * units.cm,
    )

    # close the reader
    coreas_h5_reader.close()

    return inp_file, list_file, reas_file


def write_simulation_to_file(
    inp_file, list_file, reas_file, sim_directory="./", reset=False
):
    """
    Write the provided contents of the INP, LIST and REAS files for a single simulation to the files
    in a directory named after the simulation number.

    The directory will be created in the `sim_directory` path, and will be named `SIMXXXXXX`, where
    the last six digits are the simulation number. This is retrieved from the INP file contents, as
    the first line of the file. The function will fail if the directory already exists, unless
    `reset` is set to True, in which case the directory will be removed and re-created.

    Parameters
    ----------
    inp_file : list
        The contents of the INP file
    list_file : list
        The contents of the LIST file
    reas_file : list
        The contents of the REAS file
    sim_directory : str, default='./'
        The path to the directory where the simulation directory will be made, in which all files are written
    reset : bool, default=False
        whether to delete the simulation directory or not
    """
    sim_nr = int(inp_file[0].split(" ")[-1][:-1])

    sim_directory = os.path.join(sim_directory, f"SIM{sim_nr:06d}")

    if os.path.exists(sim_directory):
        if reset:
            logger.warning(
                f"Removing and re-making simulation directory {sim_directory}"
            )
            shutil.rmtree(sim_directory)
            os.mkdir(sim_directory)
        else:
            raise FileExistsError(
                "Directory for simulation already exists! Aborting..."
            )
    else:
        os.mkdir(sim_directory)

    with open(f"{sim_directory}/SIM{sim_nr:06d}.inp", "w+") as file:
        file.writelines(inp_file)

    with open(f"{sim_directory}/SIM{sim_nr:06d}.list", "w+") as file:
        file.writelines(list_file)

    with open(f"{sim_directory}/SIM{sim_nr:06d}.reas", "w+") as file:
        file.writelines(reas_file)


# Function to sample input parameters


def sample_primary_energy(exp_range=(8, 10), size=1):
    r"""
    Sample a single primary energy from log-uniform distribution, between the exponents given
    by `exp_range`. For example, using the default settings, calling this function will generate
    a value between :math:`10^{8}` and :math:`10^{10}` GeV.

    Parameters
    ----------
    exp_range : tuple, default=(8, 10)
        The lower and upper exponent of the energy range in GeV to sample from
    size : int, default=1
        The number of energies to retrieve

    Returns
    -------
    primary_energy : list of float
        Primary energy (in internal units), as an array of length `size`
    """
    primary_energy_exp = np.random.uniform(
        *exp_range, size=size
    )  # random exponent in GeV

    return 10**primary_energy_exp * units.GeV


def sample_zenith_angle(range=(0, 90), size=1, uniformity="sin2"):
    r"""
    Sample the zenith angle from some distribution normalised with the solid angle.
    The names of the distributions refer to the variable in which they are uniform.

    Parameters
    ----------
    range : tuple, default=(0, 90)
        The lower and upper zenith angle in degrees (endpoints exclude
    size : int, default=1
        The number of angles to samples
    uniformity : {'sin2', 'sin', 'cos'}
        The distribution to use to sample the zenith angle

    Returns
    -------
    zenith_angle : list of float
        Sampled zenith angles (in internal units), as an array of length `size`

    Notes
    -----
    The names of the distributions refer to the variable in which they are uniform,
    in the sense that if you sampled using 'sin2' (the default), the zenith angle
    distribution will look uniform if binned in :math:`sin^2(\theta)`.
    """
    rng = np.random.default_rng()

    possible_theta = np.arange(*range, 0.001)[
        1:-2
    ]  # exclude 0 and 90 degrees to avoid problems with sin and division
    possible_theta *= units.deg

    if uniformity == "sin":
        # Probability distribution = sin(theta), normalised
        my_p = np.cos(possible_theta)
        my_p /= np.sum(my_p)
    elif uniformity == "sin2":
        my_p = np.sin(2 * possible_theta)
        my_p /= np.sum(my_p)
    elif uniformity == "cos-1":
        # return np.arccos(1 - rng.random(size=size))
        my_p = np.sin(possible_theta)
        my_p /= np.sum(my_p)

    return rng.choice(possible_theta, p=my_p, size=size) * units.deg


def sample_azimuth_angle(range=(0, 360), size=1):
    """
    Sample the azimuth angle within the given range, from a uniform distribution.

    Parameters
    ----------
    range : tuple, default=(0, 360)
        The lower and upper azimuth angle in degrees
    size : int, default=1
        The number of angles to samples

    Returns
    -------
    azimuth_angle : list of float
        Sampled azimuth angles (in internal units), as an array of length `size`
    """
    rng = np.random.default_rng()

    return rng.uniform(*range, size=size) * units.deg


def get_first_height(
    mass_type: str,
    xmax: float,
    log10E: float,
    zenith: float,
    atm: Atmosphere,
    obs_lvl: float = 0.0 * units.m,
    hadr_model: str = "sibyll",
    site: str = "lofar",
) -> float:
    """
    Get the first interaction height given Xmax, zenith angle, and energy.

    Note: the fit parameters used to compute the first interaction height
    are stored in a local directory 'first_height_parameters'.
    It is assumed that the file path is preserved.

    Parameters
    ----------
    mass_type : str
        The mass type of the primary particle. One of {'proton', 'iron'}.
        Other particles are not supported as of now.
    xmax : float or np.ndarray
        The depth of shower maximum in g/cm².
    log10E : float
        The logarithm (base 10) of the primary energy in eV.
    zenith : float
        The zenith angle in radians.
    atm : Atmosphere
        The Atmosphere object to use for the height conversion.
    obs_lvl : float, default=0.0 * units.m
        the observation level in meters.
    hadr_model : str, default='sibyll'
        The hadronic interaction model to use. One of {'sibyll', 'qgsjetII'}.
        This is necessary as the parameter dictionary read from the file
        (can be) hadronic interaction model dependent.
    site : str, default='lofar'
        The site to use for the parameterisation. One of {'lofar', 'auger'}.

    Returns
    -------
    first_height : float
        The first interaction height in meters.
    """
    # import the fit parameters
    fit_param_file = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "first_height_parameters",
        f"first_height_fit_params_{hadr_model}_{site}.json",
    )
    with open(fit_param_file) as f:
        fit_param_dict = json.load(f)

    # check if mass type is supported
    assert mass_type in fit_param_dict, f"Mass type {mass_type} not supported!"

    # check the value of xmax, if its larger than some range then raise a warning
    if (xmax < xmax_min):
        logger.warning(
            f"xmax value {xmax} g/cm^2 is lower than 400 g/cm^2, beyond the fit range! Results may be unreliable."
        )
    elif (xmax > xmax_max):
        logger.warning(
            f"xmax value {xmax} g/cm^2 is larger than 1300 g/cm^2, beyond the fit range! Results may be unreliable."
        )
    # compute the global fit results
    slope_params = fit_param_dict[mass_type]["slope"]
    intercept_params = fit_param_dict[mass_type]["intercept"]
    first_depth = (
        slope_params["mean_slope"] * log10E + slope_params["mean_intercept"]
    ) * xmax + (
        intercept_params["mean_slope"] * log10E + intercept_params["mean_intercept"]
    )

    # transform into height
    first_height = atm.get_vertical_height(
        zenith=zenith,
        xmax=first_depth,
        observation_level=obs_lvl / units.m,
    ) * units.m

    # if it exceeds the maximum height of the atmosphere model, set it to that
    if first_height > h_max:
        logger.warning(
            f"First interaction height {first_height/units.km:.2f} km exceeds maximum height of atmosphere model ({h_max/units.km:.2f} km). Setting to {h_max/units.m - 1:.2f} m."
        )
        first_height = h_max - 1.0 * units.m # shift slightly so that CORSIKA accepts the value

    return first_height



import argparse
import os.path

import numpy as np
import matplotlib.pyplot as plt

from smiet import units
from smiet.numpy import SlicedShower, TemplateSynthesis


## Plotting settings
plt.style.use("seaborn-v0_8-paper")
plt.rc("font", size=10)  # controls default text size
plt.rc("font", **{"family": "serif", "serif": ["Computer Modern"]})
plt.rc("axes", titlesize=12)  # fontsize of the title
plt.rc("axes", labelsize=16)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=12)  # fontsize of the x ticks
plt.rc("ytick", labelsize=12)  # fontsize of the y ticks
plt.rc("text", usetex=True)


## Create some functions for later
def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Simple example of using template synthesis, comparing to CoREAS"
    )

    parser.add_argument(
        "--frequencies",
        "-f",
        type=float,
        nargs=3,
        default=[30, 500, 100],
        help="The frequency range to use for synthesis",
    )
    parser.add_argument(
        "--ant",
        type=int,
        default=1,
        help="The index of the antenna to plot, in the synthesis arrays",
    )

    parser.add_argument(
        "--origin",
        type=str,
        default="SIM145024.hdf5",
        help="Path to the origin shower",
    )

    parser.add_argument(
        "--target",
        type=str,
        default="SIM145060.hdf5",
        help="Path to the target shower",
    )

    return parser.parse_args()


def get_target_traces(target, ant_name_select, freq_range):
    geo_target, ce_target, start_target = target.get_trace(
        ant_name_select, return_start_time=True
    )
    geo_target_filtered = target.filter_trace(np.sum(geo_target, axis=0), *freq_range)
    ce_target_filtered = target.filter_trace(np.sum(ce_target, axis=0), *freq_range)

    return geo_target_filtered, ce_target_filtered, start_target[0]


def plot_synth_coreas_comparison(
    synthesis: TemplateSynthesis,
    target: SlicedShower,
    geo_synth: np.ndarray,
    geo_target_filtered: np.ndarray,
    ce_synth: np.ndarray,
    ce_target_filtered: np.ndarray,
    ant_select,
):
    # Time axis for the synthesised traces
    time_axis = synthesis.get_time_axis()

    # Time axis for the target shower
    sampling_time = target.coreas_settings["time_resolution"]
    target_time_axis = (
        start_target + np.arange(len(geo_target_filtered)) * sampling_time
    )

    # Plotting
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))

    ax[0].plot(
        target_time_axis / units.ns,
        geo_target_filtered / (units.mV / units.m),
        c="k",
        alpha=0.5,
    )
    ax[0].plot(
        time_axis[ant_select] / units.ns,
        geo_synth[ant_select] / (units.mV / units.m),
        c="magenta",
    )

    ax[1].plot(
        target_time_axis / units.ns,
        ce_target_filtered / (units.mV / units.m),
        c="k",
        alpha=0.5,
    )
    ax[1].plot(
        time_axis[ant_select] / units.ns,
        ce_synth[ant_select] / (units.mV / units.m),
        c="magenta",
    )

    ax[0].set_xlim([target_time_axis[0], target_time_axis[200]])
    ax[1].set_xlim([target_time_axis[0], target_time_axis[200]])

    ax[0].set_xlabel("Time [ns]")
    ax[1].set_xlabel("Time [ns]")

    ax[0].set_ylabel("Geomagnetic trace [mV / m]")
    ax[1].set_ylabel("Charge-excess trace [mV / m]")

    return fig


if __name__ == "__main__":
    args = parse_arguments()

    # Check if the showers exist
    if not os.path.exists(args.origin) or not os.path.exists(args.target):
        raise FileNotFoundError(
            f"Either {args.origin} or {args.target} could not be found. "
            f"Did you download them using download_origin_showers.sh?"
        )

    ## Load in the showers
    # Both are SlicedShowers in this case, such that we can easily compare the synthesis result to CoREAS
    # But technically the target only needs to be a Shower object
    origin = SlicedShower(args.origin)
    target = SlicedShower(args.target)

    ## Do the synthesis
    # Create the synthesis object
    synthesis = TemplateSynthesis(freq_ar=list(args.frequencies))

    # Make the template from the origin - optionally set `ce_lin=False` to use quadratic term for CE
    synthesis.make_template(origin)  # process into template

    # Map the template onto the target -> output shape = (ant, samples)
    geo, ce = synthesis.map_template(target)  # map onto the target shower

    ## analysis
    # The synthesis is done! Now we can use the results in our analysis
    # Here we just show a plot to compare synth to CoREAS

    # Retrieve the traces from the target shower -> shape (slices, samples)
    ant_names = synthesis.get_antenna_names()
    ant_name_select = ant_names[args.ant]

    ant_pos_select = synthesis.antenna_information["position_showerplane"][
        synthesis.antenna_information["name"] == ant_name_select
    ]
    geo_target_filtered, ce_target_filtered, start_target = get_target_traces(
        target, ant_name_select, synthesis.frequency_range[:2]
    )

    figure = plot_synth_coreas_comparison(
        synthesis, target, geo, geo_target_filtered, ce, ce_target_filtered, args.ant
    )

    figure.suptitle(
        f"Synthesis from {synthesis.template_information['xmax']:.2f} g/cm2 to {target.xmax:.2f} g/cm2 \n Antenna is {ant_name_select} at {np.linalg.norm(ant_pos_select):.1f} m in shower plane",
        size=20,
    )

    figure.savefig("simple_example.pdf", dpi=250, bbox_inches='tight')



import logging
from typing_extensions import Self

import jax

jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp
import numpy as np

from smiet import units
from jax_radio_tools import hp_spherical_to_cartesian, cstrafo
from ..utilities.geometry import angle_between
from ..utilities.jax_utils import ModelMeta

logger = logging.getLogger("smiet.jax.io")


class BaseShower(metaclass=ModelMeta):
    """
    Base class to contain the shower parameters.

    The Shower class is used to hold the geometric information for a shower, like the zenith/azimuth,
    as well as the longitudinal profile. It can be used to specify the target parameters for a shower
    in template synthesis.
    """

    __parameter_names = ("xmax", "nmax", "zenith", "azimuth", "magnetic_field")

    def __init__(self: Self) -> None:
        self.logger = logging.getLogger("smiet.jax.io.Shower")

        # shower parameters
        self.__xmax = None
        self.__nmax = None
        self.__grammages = None
        self.__long_profile = None
        self.__GH_parameters = None
        self.__transformer = None

        # observational parameters
        self.__zenith = None
        self.__azimuth = None
        self.__geometry = None  # not used, there for backwards compatibility
        self.__magnetic_field = None
        self.__simulation_core = jnp.array([0, 0, 0])  # Currently fixed

    @property
    def xmax(self: Self) -> float:
        """The $X_{max}$ of the shower, from fitting a GH to the longitudinal profile."""
        if self.__xmax is not None:
            return self.__xmax
        else:
            self.logger.error("Xmax has not been set")

    @property
    def nmax(self: Self) -> float:
        """The $N_{max}$ of the shower, from fitting a GH to the longitudinal profile."""
        if self.__nmax is not None:
            return self.__nmax
        else:
            self.logger.error("Nmax has not been set")

    @property
    def grammages(self: Self) -> jax.Array:
        """Array of atmospheric slices in g/cm^2."""
        if self.__grammages is not None:
            return self.__grammages
        else:
            self.logger.error("The grammage has not been set")

    @property
    def slice_grammage(self: Self) -> float:
        """The spacing between each grammage in g/cm^2."""
        if self.__grammages is not None:
            return self.__grammages[1] - self.__grammages[0]
        else:
            self.logger.error(
                "Longitudinal profile has not been set yet, cannot calculate slicing thickness"
            )

    @property
    def long_profile(self: Self) -> float:
        """The longitudinal profile of electrons and positrons in each atmospheric slice."""
        if self.__long_profile is not None:
            # ensure that the length of grammages and ___long_profile is the same
            assert len(self.__long_profile) == len(self.__grammages), (
                "Length not same as grammages!"
            )
            return self.__long_profile
        else:
            self.logger.error("Set the longitudinal parameters first.")

    @property
    def GH_parameters(self: Self) -> dict:
        """Get the Gaisser-Hillas parameters as a dictionary."""
        if self.__GH_parameters is not None:
            return self.__GH_parameters
        else:
            self.logger.error("GH parameters have not been set")

    @property
    def nr_slices(self: Self) -> int:
        """The number of slices in the array."""
        if self.__long_profile is not None or self.__grammages is not None:
            # ensure that the length of grammages and __long_profile is the same
            assert len(self.__long_profile) == len(self.__grammages), (
                "Length not same as grammages!"
            )
            return len(self.__long_profile)
        else:
            self.logger.error("Set the longitudinal parameters first.")

    @property
    def geometry(self: Self) -> jax.Array:
        """Store the zenith and azimuth. These must be provided in the internal units system."""
        if self.__geometry is not None:
            return self.__geometry
        else:
            self.logger.error("Geometry has not been set")

    @property
    def zenith(self: Self) -> float:
        """The zenith angle in radians."""
        if self.__zenith is not None:
            return self.__zenith
        else:
            self.logger.error("Geometry has not been set")

    @property
    def azimuth(self: Self) -> float:
        """The azimuthal angle in radians."""
        if self.__azimuth is not None:
            return self.__azimuth
        else:
            self.logger.error("Geometry has not been set")

    @property
    def core(self: Self) -> jax.Array:
        """The core (x, y, z in NRR CS) where the EAS hit in the simulation."""
        if self.__simulation_core is not None:
            return self.__simulation_core
        else:
            self.logger.error("The simulation core is not known")

    @property
    def magnet(self: Self) -> float:
        """Magnetic field vector in the NRR coordinate system."""
        if self.__magnetic_field is not None:
            return self.__magnetic_field
        else:
            self.logger.error("The magnetic field vector has not been set")

    @property
    def geomagnetic_angle(self: Self) -> float:
        """The angle between the magnetic field vector and the shower axis."""
        shower_axis = hp_spherical_to_cartesian(
            self.__zenith / units.rad, self.__azimuth / units.rad
        )

        return angle_between(self.__magnetic_field, shower_axis)

    @xmax.setter
    def xmax(self: Self, xmax: float) -> None:
        self.__xmax = xmax

    @nmax.setter
    def nmax(self: Self, nmax: float) -> None:
        self.__nmax = nmax

    @zenith.setter
    def zenith(self: Self, zenith: float) -> None:
        self.__zenith = zenith

    @azimuth.setter
    def azimuth(self: Self, azimuth: float) -> None:
        self.__azimuth = azimuth

    @geometry.setter
    def geometry(self: Self, geo: jax.typing.ArrayLike) -> None:
        assert len(geo) == 2, (
            "Please provide zenith and azimuth components in internal units"
        )

        self.__geometry = jnp.array(geo)

    @grammages.setter
    def grammages(self: Self, grammages: jax.typing.ArrayLike) -> None:
        self.__grammages = grammages

    @long_profile.setter
    def long_profile(self: Self, long_profile: jax.typing.ArrayLike) -> None:
        self.__long_profile = long_profile

    @GH_parameters.setter
    def GH_parameters(self: Self, params: dict) -> None:
        self.__GH_parameters = params

    @magnet.setter
    def magnet(self: Self, magnet_field_vector: jax.typing.ArrayLike) -> None:
        assert len(magnet_field_vector) == 3, (
            "B-field vector must contain three components"
        )

        self.__magnetic_field = magnet_field_vector

    @core.setter
    def core(self: Self, core: jax.typing.ArrayLike) -> None:
        self.__simulation_core = core

    def get_transformer(self: Self):
        """
        Get the transformer from jax_radio_tools that contains the coordinate transformations from shower to ground frame, as well to on-sky coordinates.

        Returns
        -------
        transformer : cstrafo
            The coordinate transformer from jax_radio_tools.
        """
        if self.__transformer is None:
            self.__transformer = cstrafo(
                zenith=self.zenith / units.rad,
                azimuth=self.azimuth / units.rad,
                magnetic_field_vector=self.magnet,
            )

        return self.__transformer

    def set_parameters(
        self: Self,
        grammages: jax.typing.ArrayLike,
        long_profile: jax.typing.ArrayLike,
        xmax : jax.typing.ArrayLike,
        nmax : jax.typing.ArrayLike,
        zenith : jax.typing.ArrayLike,
        azimuth : jax.typing.ArrayLike,
        magnetic_field : jax.typing.ArrayLike,
        core : jax.typing.ArrayLike,
    ) -> None:
        """
        Set the parameters of the shower model from a dictionary of parameters.

        This is a convenience function to set all parameters at once.

        Parameters
        ----------
        grammages : jax.typing.ArrayLike
            an array of atmospheric depth in g/cm^2
        long_profile : jax.typing.ArrayLike
            an array of the longitudinal profile of the shower
        xmax : jax.typing.ArrayLike
            depth of shower maximum in g/cm^2
        nmax : jax.typing.ArrayLike
            number of particles at shower maximum
        zenith : jax.typing.ArrayLike
            zenith angle in radians
        azimuth : jax.typing.ArrayLike
            azimuthal angle in radians
        magnetic_field : jax.typing.ArrayLike
            magnetic field vector in Gauss. Must be a 3-vector
        core : jax.typing.ArrayLike
            core position in meters. Must be a 3-vector
        """
        self.grammages = grammages
        self.long_profile = long_profile

        self.xmax = xmax
        self.nmax = nmax

        self.zenith = zenith
        self.azimuth = azimuth
        self.magnet = magnetic_field
        self.core = core

    def transform_profile_to_origin(
        self: Self, origin_grammages: jax.typing.ArrayLike
    ) -> None:
        """
        Transform the longitudinal profile to match those of the origin shower.

        This is needed such that the grammages of the origin and target match.

        Parameters
        ----------
        origin_grammages : jax.typing.ArrayLike
            The grammages of the origin shower.
        """
        self.long_profile = jnp.interp(
            x=origin_grammages, xp=self.grammages, fp=self.long_profile
        )
        self.grammages = origin_grammages


import jax
import jax.numpy as jnp

# Core interpolation with boundary handling
def linear_interp_with_bounds(x, xp, fp):
    # Find the right index for interpolation
    idx = jnp.searchsorted(xp, x, side='right') - 1
    idx = jnp.clip(idx, 0, len(xp) - 2)

    # Interpolation logic
    x0 = jnp.take(xp, idx)
    x1 = jnp.take(xp, idx + 1)
    y0 = jnp.take(fp, idx)
    y1 = jnp.take(fp, idx + 1)

    slope = (y1 - y0) / (x1 - x0)
    y = y0 + slope * (x - x0)

    # Clamp at boundaries
    y = jnp.where(x < xp[0], fp[0], y)
    y = jnp.where(x > xp[-1], fp[-1], y)
    return y


def batched_interp1d(x, xp, fp):
    """
    Batched linear interpolation with boundary handling.

    For left and right values, the values at the boundaries are used.

    Parameters
    ----------
    x : jnp.ndarray
        Input points to interpolate. Shape: (Nant, Nslice).
    xp : jnp.ndarray
        Points to interpolate at. Shape: (Ngrid,).
    fp : jnp.ndarray
        Function values at xp. Shape: (Ngrid, Nfreq, Nslice).

    Returns
    -------
    jnp.ndarray
        Interpolated values. Shape: (Nant, Nfreq, Nslice).
    """
    Nant, Nslice = x.shape
    Ngrid, Nfreq, _ = fp.shape

    # Function to interpolate over 1 slice and 1 freq
    def interp_fn(f, s):
        x_query = x[:, s]           # shape: (Nant,)
        values = fp[:, f, s]        # shape: (Ngrid,)
        return jax.vmap(linear_interp_with_bounds, in_axes=(0, None, None))(x_query, xp, values)  # → (Nant,)

    # Generate all (f, s) index pairs
    freq_idx = jnp.arange(Nfreq)
    slice_idx = jnp.arange(Nslice)

    # Vectorize over frequency and slice
    interp_fs = jax.vmap(
        lambda f: jax.vmap(
            lambda s: interp_fn(f, s)
        )(slice_idx)
    )(freq_idx)  # shape: (Nfreq, Nslice, Nant)

    # Reorder to (Nant, Nfreq, Nslice)
    return jnp.transpose(interp_fs, (2, 0, 1))


import logging
from typing_extensions import Self, Union, Tuple

import h5py
import numpy as np

import radiotools.atmosphere.models as models

from .base_shower import Shower, CoreasHDF5
from smiet import units
from smiet.numpy.utilities import bandpass_filter_trace, e_to_geo_ce


class CoreasShower(Shower, CoreasHDF5):
    """
    A class to read in a CoREAS shower from an HDF5 file.

    Whereas the :class:`smiet.numpy.io.CoreasHDF5` class provides the basic functionality to
    read in the settings and the shower data from a CoREAS HDF5 file, this class extends it to provide
    with the functionality from the :class:`smiet.numpy.io.Shower` class. This allows you to
    interact with the antennas more easily, as the positions are more readily accessible. Also, the
    geometry is included such that a coordinate transformation object can be created from the shower.

    Parameters
    ----------
    file_path : str
        The path to the CoREAS HDF5 file
    gdas_file : str, optional
        If provided, the atmosphere will be created using this GDAS file. This will take
        precedence over the atmosphere model specified in the simulation settings.
    """
    def __init__(self, file_path, gdas_file=None):
        CoreasHDF5.__init__(self, file_path)
        Shower.__init__(self)

        self.logger = logging.getLogger("smiet.numpy.io.CoreasShower")

        # Read in the shower settings from the HDF5 file
        self.magnet = self.corsika_settings['magnetic_field']
        self.zenith = self.corsika_settings['zenith']
        self.azimuth = self.corsika_settings['azimuth']
        self.core = self.coreas_settings['core_coordinate']
        self.grammages, self.long_profile = self.get_long_profile()
        self.delta_t = self.coreas_settings['time_resolution']

        # set core coordinate z-coordinate to be observation level
        if self.core[2] != self.corsika_settings['observation_level'] * units.cm:
            self.logger.warning(f"Setting core z-coordinate to be observation level at {self.corsika_settings['observation_level'] * units.cm /units.m} m.")
            self.core[2] = self.corsika_settings['observation_level'] * units.cm


        self.GH_parameters = self.get_gaisser_hillas_parameters()

        # Check if we can create the atmosphere
        if gdas_file is not None:
            self.logger.info(f"Creating non-curved Atmosphere object with GDAS file {gdas_file}")
            self.atmosphere = models.Atmosphere(
                gdas_file=gdas_file,
                curved=False
            )
        elif self.corsika_settings['atmosphere_model'] in models.atm_models.keys():
            self.logger.info(
                f"Creating non-curved Atmosphere object with model {self.corsika_settings['atmosphere_model']} and"
                f"refractive index at sea level n0={self.coreas_settings['n0']}"
            )
            self.atmosphere = models.Atmosphere(
                model = self.corsika_settings['atmosphere_model'],
                n0 = self.coreas_settings['n0'],
                curved=False
            )
        else:
            self.logger.error(
                f"Atmosphere model {self.corsika_settings['atmosphere_model']} not supported. "
                "Please provide a valid atmosphere model or a GDAS file."
            )

        # To be assigned later
        self.antenna_names = None
        self.antenna_array = None
        self._trace_length = None

        with h5py.File(self._file) as file:
            self._read_antennas(file)

    def _read_antennas(self, file):
        # If the antenna names have already been set (in child class for example),
        # this function should not be executed
        if self.antenna_names is not None:
            return

        self.logger.info(f"Reading antennas from {file.name}...")

        self.antenna_names = set(file["CoREAS/observers"])

        self._trace_length = len(
            file["CoREAS"]["observers"][
                f"{next(iter(self.antenna_names))}"
            ]
        )

        # Create the antenna array, which is a structured array containing the name and position of each antenna
        self._make_antenna_array({
            ant_name: file["CoREAS/observers"][ant_name].attrs["position"] * units.cm
            for ant_name in self.antenna_names
        })

        # And sort it, such that the order is always the same
        self._sort_antenna_array()

    def _make_antenna_array(self, observers_pos: dict):
        self.antenna_array = np.zeros(
            len(self.antenna_names),
            dtype=np.dtype([
                ('name', 'U20'),
                ('position', 'f8', 3)
            ])
        )
        for idx, ant in enumerate(observers_pos.items()):
            self.antenna_array[idx] = (
                ant[0],
                ant[1],
            )

        # Convert positions to our CS
        self.antenna_array['position'] = self.antenna_array['position'][:, [1, 0, 2]]
        self.antenna_array['position'][:, 0] *= -1  # west -> east

    def _sort_antenna_array(self):
        """
        Sort the antenna array by distance to the core.

        This function ensures that the order of the ``antenna_array`` is always the same when reading in
        a particular shower.
        """
        distance_to_core = np.linalg.norm(
            (self.core - self.antenna_array['position']), axis=-1
        )

        first_axis = self.antenna_array['position'][:, 0]

        # Round the distance_to_core to be sure to recognise the rings of starshape,
        # even with small numerical errors. Sort by distance, using first axis as
        # the tie-breaker
        self.antenna_array = self.antenna_array[np.lexsort((first_axis, np.round(distance_to_core, 2)))]

    def __find_antenna_index(self, ant_names):
        """
        Find the index of the antenna with name `ant_name` in the antenna array.

        Parameters
        ----------
        ant_names : str or list of str
            Either a single antenna names or a list of antenna names for which to retrieve the positions.
            All names must be present in ``SlicedShower.antenna_names``!

        Returns
        -------
        antenna_index : int
            The index or indices of the antenna with name `ant_name` in the antenna array.
        """
        if isinstance(ant_names, (str, np.str_)):
            antenna_idx = np.nonzero(self.antenna_array['name'] == ant_names)[0][0]
        else:
            antenna_idx = [np.nonzero(self.antenna_array['name'] == ant_name)[0][0] for ant_name in ant_names]
        return antenna_idx

    def get_antenna_position(self, ant_names):
        """
        Get the position of the antennas in `ant_names` on the ground.

        The position is returned in our coordinate system, where the x-axis points to the magnetic east,
        the y-axis to the magnetic north and the z-axis up.

        Parameters
        ----------
        ant_names : str or list of str
            Either a single antenna names or a list of antenna names for which to retrieve the positions.
            All names must be present in ``SlicedShower.antenna_names``!

        Returns
        -------
        antenna_ground : np.ndarray
            The positions of the antennas in `ant_names`.
        """
        antenna_idx = self.__find_antenna_index(ant_names)
        antenna_ground = self.antenna_array['position'][antenna_idx]
        return antenna_ground

    def get_antenna_position_showerplane(self, ant_names):
        r"""
        Get the position of the antennas in `ant_names` in the shower plane.

        The shower plane coordinate system is that from ``radiotools``,
        so the x-axis points along the :math:`\vec{v} \times \vec{B}` direction,
        the y-axis along :math:`\vec{v} \times ( \vec{v} \times \vec{B} )`
        direction and the z-axis along :math:`\vec{v}`.

        Parameters
        ----------
        ant_names : str or list of str
            Either a single antenna names or a list of antenna names for which to retrieve the positions.
            All names must be present in ``SlicedShower.antenna_names``!

        Returns
        -------
        antenna_showerplane : np.ndarray
            The position of the antennas in the shower plane.
        """
        transformer = self.get_transformer()

        antenna_ground = self.get_antenna_position(ant_names)  # shape (n_antennas, 3)
        antenna_showerplane = transformer.transform_to_vxB_vxvxB(antenna_ground, core=self.core)

        return antenna_showerplane
    
    def get_trace_geoce(
        self: Self,
        ant_name: str,
        bandpass: list = [30 * units.MHz, 500 * units.MHz],
        return_times: bool = False,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]:
        """
        Get the traces for a given antenna, in GEO/CE components.

        Notes
        -----
        The traces are converted to GEO/CE components.

        Parameters
        ----------
        ant_name : str
            The name of the antenna. Must be the same as the key in the HDF5!
        bandpass : list, default=[30, 500]
            If not None, the traces are bandpass filtered between the given frequencies (in MHz)
        return_times : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        traces_geo : np.ndarray
            The geomagnetic traces, shaped as (samples,)
        traces_ce : np.ndarray
            The charge-excess traces, shaped as (samples,)
        return_times : float, optional
            The time of the first sample of the traces, in seconds.
        """
        if ant_name not in self.antenna_names:
            raise ValueError(f"Antenna name {ant_name} is not present in shower")

        transformer = self.get_transformer()
        trace_slice_ground, trace_time_axis = self.get_trace_raw(ant_name
        )
        trace_slice_filtered = bandpass_filter_trace(
            trace_slice_ground,
            self.coreas_settings["time_resolution"],
            bandpass[0],
            bandpass[1],
            sample_axis=-1,
        )
        trace_slice_vvB = transformer.transform_to_vxB_vxvxB(trace_slice_filtered).T

        # unit of pos does not matter, this is divided away
        antenna_vvB = self.get_antenna_position_showerplane(ant_name)
        traces_geo, traces_ce = e_to_geo_ce(trace_slice_vvB, *antenna_vvB[:2])

        if return_times:
            return traces_geo, traces_ce, trace_time_axis

        return traces_geo, traces_ce
    
    def get_traces_geoce(
        self: Self,
        ant_names: Union[str, list],
        bandpass: list = [30 * units.MHz, 500 * units.MHz],
        return_times: bool = False,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]:
        """
        Get the traces for a given antenna, in GEO/CE components.

        Notes
        -----
        The traces are converted to GEO/CE components.

        Parameters
        ----------
        ant_names : Union[str, list]
            The name of the antenna or names of antennas. Must have the same keys in the HDF5!
        bandpass : list, default=[30, 500]
            If not None, the traces are bandpass filtered between the given frequencies (in MHz)
        return_times : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        traces_geo : np.ndarray
            The geomagnetic traces, shaped as (n_antennas, samples,)
        traces_ce : np.ndarray
            The charge-excess traces, shaped as (n_antennas, samples,)
        traces_start_times : shaped as n_antennas, optional
            The time of the first sample of the traces, in seconds.
        """
        traces_geo = []
        traces_ce = []
        trace_times = []

        if isinstance(ant_names, (str, np.str_)):
            ant_names = [ant_names]
        for ant_name in ant_names:
            trace_geo, trace_ce, trace_time_axis = self.get_trace_geoce(
                ant_name,
                bandpass=bandpass,
                return_times=True,
            )
            traces_geo.append(trace_geo)
            traces_ce.append(trace_ce)
            trace_times.append(trace_time_axis)

        traces_geo = np.array(traces_geo)
        traces_ce =  np.array(traces_ce)
        if return_times:
            return traces_geo, traces_ce, np.array(trace_times)

        return traces_geo, traces_ce
    
    def get_traces_vB_vvB(
        self: Self,
        ant_names: Union[str, list],
        bandpass: list = [30 * units.MHz, 500 * units.MHz],
        return_times: bool = False,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]:
        """
        Get the traces for a given antenna, in v x B and v x (v x B) components.

        Parameters
        ----------
        ant_names : Union[str, list]
            The name of the antenna or names of antennas. Must have the same keys in the HDF5!
        bandpass : list, default=[30, 500]
            If not None, the traces are bandpass filtered between the given frequencies (in MHz)
        return_times : bool, default=False
            If True, an array containing the time of the first sample of each slice is returned

        Returns
        -------
        traces_vxB : np.ndarray
            The v x B traces, shaped as (n_antennas, samples,)
        traces_vxvxB : np.ndarray
            The v x (v x B) traces, shaped as (n_antennas, samples,)
        traces_start_times : shaped as n_antennas, optional
            The time of the first sample of the traces, in seconds.
        """
        traces_vxB = []
        traces_vxvxB = []
        trace_times = []

        if isinstance(ant_names, (str, np.str_)):
            ant_names = [ant_names]
        for ant_name in ant_names:
            transformer = self.get_transformer()
            trace_slice_ground, trace_time_axis = self.get_trace_raw(ant_name
            )
            trace_slice_filtered = bandpass_filter_trace(
                trace_slice_ground,
                self.coreas_settings["time_resolution"],
                bandpass[0],
                bandpass[1],
                sample_axis=-1,
            )
            trace_slice_vvB = transformer.transform_to_vxB_vxvxB(trace_slice_filtered)

            traces_vxB.append(trace_slice_vvB[0])
            traces_vxvxB.append(trace_slice_vvB[1])
            trace_times.append(trace_time_axis)

        traces_vxB = np.array(traces_vxB)
        traces_vxvxB =  np.array(traces_vxvxB)
        if return_times:
            return traces_vxB, traces_vxvxB, np.array(trace_times)
        
        return traces_vxB, traces_vxvxB


from .cs_transformations import geo_ce_to_e, e_to_geo_ce
from .geometry import angle_between, unit_vector
from .trace_utils import (
    bandpass_filter_trace,
    transform_traces_on_vxB,
    reconstruct_starshape_from_vxvxB,
)

__all__ = [
    "geo_ce_to_e",
    "e_to_geo_ce",
    "angle_between",
    "unit_vector",
    "bandpass_filter_trace",
]



import time

import numpy as np
import matplotlib.pyplot as plt

import jax
jax.config.update('jax_platform_name', 'cpu')

from smiet import units
from smiet.jax import SlicedShower, TemplateSynthesis


# Selections (to be put in main)
FREQ = [30, 80, 50]

# Variables
# shower_path = '/user/mitjadesmet/CrData/Showers50'
# shower_path = '/home/mitjadesmet/Data/Showers/'
shower_path = '/home/kwatanabe/Projects/radio-ift/resources/template_synthesis/OriginShowers'
# shower_path = "/home/mdesmet/Data/Showers50"
shower_origin = 130165
shower_target = 130188

origin = SlicedShower(f'{shower_path}/SIM{shower_origin}.hdf5')
target = SlicedShower(f'{shower_path}/SIM{shower_target}.hdf5')

time0 = time.time()

synthesis = TemplateSynthesis(freq_ar=FREQ)

time1 = time.time()

synthesis.make_template(origin)

time2 = time.time()

total_geo_synth, total_ce_synth = synthesis.map_template(target)

time3 = time.time()

print(f'Spent {time1 - time0:.4f} initialising, \n'
      f'{time2 - time1:.4f} making template and \n'
      f'{time3 - time2:.4f} mapping it')

# Get the target trace for comparison
ant_names = synthesis.get_ant_names()
origin_time = synthesis.get_time_axis()

dt_res = target.get_coreas_settings()['time_resolution']

geo_origin, ce_origin, start_origin = origin.get_traces(return_start_time=True)
geo_target, ce_target, start_target = target.get_traces(return_start_time=True)

# PLOTTING
plt.style.use('seaborn-v0_8-paper')
plt.rc('font', size=10)  # controls default text size
plt.rc('axes', titlesize=12)  # fontsize of the title
plt.rc('axes', labelsize=16)  # fontsize of the x and y labels
plt.rc('xtick', labelsize=12)  # fontsize of the x ticks
plt.rc('ytick', labelsize=12)  # fontsize of the y ticks
# plt.rc("text", usetex=True)

for ant_idx, ant_name in enumerate(ant_names):
    # Filter target trace
    trace_filtered_geo = np.sum(
        target.filter_trace(
            geo_target[ant_idx,:], *synthesis.frequency_range[:2]
        ),
        axis=1
    )

    trace_filtered_ce = np.sum(
        target.filter_trace(
            ce_target[ant_idx,:], *synthesis.frequency_range[:2]
        ),
        axis=1
    )

    fig, ax = plt.subplots(1, 2, figsize=(12, 6))
    ax = ax.flatten()

    ax[0].plot(
        np.arange(len(trace_filtered_geo)) * dt_res + start_target[ant_idx],
        trace_filtered_geo / (units.microvolt / units.m),
        c='k', label='CoREAS',
        alpha=0.5
    )
    ax[1].plot(
        np.arange(len(trace_filtered_ce)) * dt_res + start_target[ant_idx],
        trace_filtered_ce / (units.microvolt / units.m),
        c='k', label='CoREAS',
        alpha=0.5
    )

    ax[0].plot(
        origin_time[ant_idx],
        total_geo_synth[ant_idx,:] / (units.microvolt / units.m),
        '--',
        c='magenta', label='Template Synthesis',
    )
    ax[1].plot(
        origin_time[ant_idx],
        total_ce_synth[ant_idx,:] / (units.microvolt / units.m),
        '--',
        c='magenta', label='Template Synthesis',
    )

    ax[0].set_xlim([origin_time[ant_idx][20], origin_time[ant_idx][1020]])
    ax[1].set_xlim([origin_time[ant_idx][20], origin_time[ant_idx][1020]])

    ax[0].set_ylabel(r'$E [\mu \mathrm{V}/\mathrm{m}]$', size=16)
    ax[0].set_xlabel('Time [ns]')
    ax[1].set_xlabel('Time [ns]')

    ax[0].legend(fontsize=14)
    ax[1].legend(fontsize=14)

    ax[0].set_title('Geomagnetic component')
    ax[1].set_title('Charge-excess component')

    fig.suptitle(f'Signals for antenna {ant_name} \n'
                 r'$X^{\mathrm{origin}}_{\mathrm{max}}$ = ' + f' {origin.xmax:.1f} ' + r'$\mathrm{g}/\mathrm{cm}^2$' + ' - '
                 r' $X^{\mathrm{target}}_{\mathrm{max}}$ = ' + f' {target.xmax:.1f} ' + r'$\mathrm{g}/\mathrm{cm}^2$'
                 '\n', y=1.05, size=17)

    fig.savefig(f'SynthesisedPulse_JAX_{shower_origin}_{shower_target}_{ant_name}.png',
                bbox_inches='tight', dpi=200)

    plt.close(fig)

# plt.show()
print("Done")



from .io import Shower, CoreasHDF5, CoreasShower, SlicedShower, SlicedShowerCherenkov
from .synthesis import SliceSynthesis, TemplateSynthesis
from .utilities import e_to_geo_ce, geo_ce_to_e

__all__ = [
    "Shower",
    "CoreasHDF5",
    "CoreasShower",
    "SlicedShower",
    "SlicedShowerCherenkov",
    "SliceSynthesis",
    "TemplateSynthesis",
    "e_to_geo_ce",
    "geo_ce_to_e",
]



import logging

from . import units
from .core import SMIETRun

# Define public API as the fewest amount of objects required to apply template synthesis
__all__ = [
    "units", "SMIETRun"
]

# Set up smiet logger
class ColoredFormatter(logging.Formatter):
    COLOR_CODES = {
        logging.DEBUG: "\033[94m",  # Blue
        logging.INFO: "\033[92m",  # Green
        logging.WARNING: "\033[93m",  # Yellow
        logging.ERROR: "\033[91m",  # Red
        logging.CRITICAL: "\033[95m",  # Magenta
    }

    RESET_CODE = "\033[0m"

    def format(self, record):
        # Get the color code for the log level
        color_code = self.COLOR_CODES.get(record.levelno)

        # Format the message
        message = super().format(record)

        # Add the color
        message = f"{color_code}{message}{self.RESET_CODE}"
        return message


# Make StreamHandler with the colored formatter
handler = logging.StreamHandler()
handler.setFormatter(
    ColoredFormatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(funcName)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
handler.setLevel(1)

# Add handler to logger
logger = logging.getLogger("smiet")
logger.addHandler(handler)
logger.propagate = False  # do not pass message on to root logger



import os.path
import argparse
import bisect
import numpy as np
import matplotlib.pyplot as plt

from smiet import units
from smiet.numpy import SlicedShower, TemplateSynthesis


## Plotting settings
plt.style.use("seaborn-v0_8-paper")
plt.rc("font", size=10)  # controls default text size
plt.rc("font", **{"family": "serif", "serif": ["Computer Modern"]})
plt.rc("axes", titlesize=12)  # fontsize of the title
plt.rc("axes", labelsize=16)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=12)  # fontsize of the x ticks
plt.rc("ytick", labelsize=12)  # fontsize of the y ticks
plt.rc("text", usetex=True)


## Create some functions for later
def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Simple example of using template synthesis, comparing to CoREAS"
    )

    parser.add_argument(
        "--frequencies",
        "-f",
        type=float,
        nargs=3,
        default=[30, 500, 100],
        help="The frequency range to use for synthesis",
    )
    parser.add_argument(
        "--ant",
        type=int,
        default=1,
        help="The index of the antenna to plot, in the synthesis arrays",
    )

    parser.add_argument(
        "--origin",
        type=str,
        default="SIM145024.hdf5",
        help="Path to the origin shower",
    )

    parser.add_argument(
        "--target",
        type=str,
        default="SIM145060.hdf5",
        help="Path to the target shower",
    )

    return parser.parse_args()


def get_target_slice_trace(
    target: SlicedShower,
    slice_grammage: int,
    ant_name_select: str,
    freq_range: list[float],
):
    geo_target, ce_target, start_target = target.get_trace_slice(
        ant_name_select, slice_grammage, return_start_time=True
    )
    geo_target_filtered = target.filter_trace(geo_target, *freq_range)
    ce_target_filtered = target.filter_trace(ce_target, *freq_range)

    return geo_target_filtered, ce_target_filtered, start_target


def plot_slice_synthesis(ant_select: int):
    # TODO: maybe also show slice specific antenna information as example of getter?
    fig, ax = plt.subplots(2, 2, figsize=(5, 5))
    ax = ax.flatten()

    # Time axis for the synthesised traces
    time_axis = synthesis.get_time_axis()

    # Time axis for the target shower
    sampling_time = target.coreas_settings["time_resolution"]

    for i, my_slice in enumerate([500, 700, 900, 1100]):
        # Search slice index knowing that the synthesis.slices_grammage list is sorted
        # Alternative: my_slice // slice_grammage - 1
        my_slice_ind = bisect.bisect_left(synthesis.slices_grammage, my_slice)
        geo_synth, ce_synth = synthesis.slice_map_template(my_slice_ind, target)

        geo_target_filtered, ce_target_filtered, start_target = get_target_slice_trace(
            target, my_slice, ant_name_select, synthesis.frequency_range[:2]
        )

        target_time_axis = (
            start_target + np.arange(len(geo_target_filtered)) * sampling_time
        )

        geo = ax[i].plot(
            target_time_axis / units.ns,
            geo_target_filtered / (units.mV / units.m),
            c="k",
            alpha=0.5,
            label="CoREAS",
        )
        ax[i].plot(
            time_axis[ant_select] / units.ns,
            geo_synth[ant_select] / (units.mV / units.m),
            c="magenta",
            label="synthesis",
        )

        ce = ax[i].plot(
            target_time_axis / units.ns,
            ce_target_filtered / (units.mV / units.m),
            linestyle="--",
            c="k",
            alpha=0.5,
        )
        ax[i].plot(
            time_axis[ant_select] / units.ns,
            ce_synth[ant_select] / (units.mV / units.m),
            linestyle="--",
            c="magenta",
        )

        ax[i].set_xlim([target_time_axis[0], target_time_axis[200]])
        ax[i].set_xlabel("Time [ns]")
        ax[i].set_ylabel("Electric field trace [mV / m]")
        ax[i].set_title(f"Slice at {my_slice} g/cm2")

        ax[i].legend()
        # ax[i].legend([geo, ce], ["Geomagnetic", "Charge-excess"])

    return fig


if __name__ == "__main__":
    args = parse_arguments()

    # Check if the showers exist
    if not os.path.exists(args.origin) or not os.path.exists(args.target):
        raise FileNotFoundError(
            f"Either {args.origin} or {args.target} could not be found. "
            f"Did you download them using download_origin_showers.sh?"
        )

    # Load in the showers
    # Both are SlicedShowers in this case, such that we can compare per slice
    origin = SlicedShower(args.origin)
    target = SlicedShower(args.target)

    # Create the synthesis object
    synthesis = TemplateSynthesis(freq_ar=list(args.frequencies))

    # Make the template from the origin - optionally set `ce_lin=False` to use quadratic term for CE
    synthesis.make_template(origin)  # process into template

    # Get the information about the antenna we are plotting
    ant_names = synthesis.get_antenna_names()
    ant_name_select: str = ant_names[args.ant]

    ant_pos_select = synthesis.antenna_information["position_showerplane"][
        synthesis.antenna_information["name"] == ant_name_select
    ]

    # Perform and plot the synthesis for 4 different slices
    figure = plot_slice_synthesis(args.ant)

    figure.suptitle(
        f"Synthesis per slice from {synthesis.template_information['xmax']:.2f} g/cm2 to {target.xmax:.2f} g/cm2 \n Antenna is {ant_name_select} at {np.linalg.norm(ant_pos_select):.1f} m in shower plane",
        size=20,
    )

    figure.savefig("synthesis_per_slice.pdf", dpi=250, bbox_inches='tight')



import numpy as np


# source: https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
def unit_vector(vector):
    """
    Normalize a vector to unit length

    Parameters
    ----------
    vector : array_like
        The vector to normalize

    Returns
    -------
    unit : np.ndarray
        The unit vector of the input vector.

    Examples
    --------
    >>> unit_vector(np.array([1, 0, 0]))
    array([1., 0., 0.])

    >>> unit_vector(np.array([0, 4, 0]))
    array([0., 1., 0.])

    >>> unit_vector(np.array([3, 6, -2]))
    array([ 0.42857143,  0.85714286, -0.28571429])
    """
    return np.asarray(vector) / np.linalg.norm(vector, axis=0)


def angle_between(v1, v2):
    """
    Returns the angle in radians between vectors 'v1' and 'v2'

    Parameters
    ----------
    v1 : array_like
        First vector
    v2 : array_like
        Second vector

    Returns
    -------
    angle : float
        The angle (in radians) between the vectors

    Examples
    --------
    >>> angle_between((1, 0, 0), (0, 1, 0))
    1.5707963267948966

    >>> angle_between((1, 0, 0), (1, 0, 0))
    0.0

    >>> angle_between((1, 0, 0), (-1, 0, 0))
    3.141592653589793
    """
    v1_u = unit_vector(np.asarray(v1).T).T
    v2_u = unit_vector(np.asarray(v2).T).T

    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))



import numpy as np


def e_geo(traces, x, y):
    """
    Calculate the geomagnetic component from the electric field in the shower plane,
    i.e. the electric field should be in the (vxB, vxvxB, v) CS

    Parameters
    ----------
    traces : np.ndarray
        The traces in the shower plane, shaped as (samples, polarisations)
    x : float
        The antenna position along the vxB axis
    y : float
        The antenna position along the vxvxB axis

    Returns
    -------
    e_geo : np.ndarray
        The geomagnetic component of the electric field
    """
    return traces[:, 1] * x / y - traces[:, 0]


def e_ce(traces, x, y):
    """
    Calculate the charge-excess (or Askaryan) component of electric field in the shower plane,
    i.e. the electric field should be in the (vxB, vxvxB, v) CS

    Parameters
    ----------
    traces : np.ndarray
        The traces in the shower plane, shaped as (samples, polarisations)
    x : float
        The antenna position along the vxB axis
    y : float
        The antenna position along the vxvxB axis

    Returns
    -------
    e_ce : np.ndarray
        The charge-excess component of the electric field
    """
    return -traces[:, 1] * np.sqrt(x**2 + y**2) / y


def e_to_geo_ce(traces, x, y):
    """
    Decouples the electric field in the shower plane, i.e. the electric field should be in the (vxB, vxvxB, v) CS,
    into the geomagnetic and charge-excess components.

    Parameters
    ----------
    traces : np.ndarray
        The traces in the shower plane, shaped as (samples, polarisations)
    x : float
        The antenna position along the vxB axis
    y : float
        The antenna position along the vxvxB axis

    Returns
    -------
    e_geo : np.ndarray
        The geomagnetic component of the electric field
    e_ce : np.ndarray
        The charge-excess component of the electric field
    """
    return e_geo(traces, x, y), e_ce(traces, x, y)


def geo_ce_to_e(my_e_geo, my_e_ce, x, y):
    """
    Convert the geomagnetic and charge-excess components to a three-dimensional electric field in
    the shower plane, i.e. the (vxB, vxvxB, v) CS.

    Note that the v-component is simply set to zero.

    Parameters
    ----------
    my_e_geo: np.ndarray
        The electric field traces of the geomagnetic component, shaped as (n_antennas, n_samples, n_slices,)
    my_e_ce: np.ndarray
        The electric field traces of the charge-excess component, shaped as (n_antennas, n_samples, n_slices,)
    x : float
        The antenna position along the vxB axis, shaped as (n_antennas, )
    y : float
        The antenna position along the vxvxB axis, shaped as (n_antennas, )

    Returns
    -------
    e_field : np.ndarray
        The three-dimensional electric field, in the shower plane CS, shaped as (samples, polarisations)
    """
    if hasattr(x, "__len__"):
        x = np.asarray(x)
        y = np.asarray(y)
    else:
        x = np.asarray([x])
        y = np.asarray([y])

    # expand the x and y positions to match the traces
    expand_axes = (1, 2) if my_e_geo.ndim == 3 else (1,)
    x_arr = np.expand_dims(x, axis=expand_axes)
    y_arr = np.expand_dims(y, axis=expand_axes)

    trace_vB = -1 * (
        my_e_geo
        + my_e_ce
        * x_arr
        / np.sqrt(x_arr ** 2 + y_arr ** 2)
    )
    trace_vvB = (
        -1
        * my_e_ce
        * y_arr
        / np.sqrt(x_arr ** 2 + y_arr ** 2)
    )
    trace_v = np.zeros_like(trace_vB)

    return np.stack((trace_vB, trace_vvB, trace_v), axis=-1)



"""Utility module to take care of transformations to shower plane etc."""

import jax

jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp
from jax_radio_tools import geo_ce_to_vB_vvB_v


def transform_traces_on_vxB(
    traces_geoce: jax.typing.ArrayLike,
    x: jax.typing.ArrayLike,
    y: jax.typing.ArrayLike,
    vxB_axis_mode: str = "replace",
) -> jax.Array:
    """
    Adjust the (synthesised) traces on the vxB axis.

    This function is meant to alter the traces on the vxB axis, as in the current
    version of the SMIET software the vxB axis cannot be synthesised. There are several
    strategies to handle this, and applying them should be facilitated by this function.

    Here it is assumed that the antennas are ordered with increasing phi, where phi=0 is oriented West.

    Parameters
    ----------
    traces_geoce : jax.typing.ArrayLike
        the synthesised traces in geomagnetic and charge-excess polarisation.
        Shape must be of NPOL x NANTS x NSAMPLES
    x : jax.typing.ArrayLike
        the antenna positions along the vxB axis, shaped as (NANTS,)
    y : jax.typing.ArrayLike
        the antenna positions along the vxvxB axis, shaped as (NANTS,)
    vxB_axis_mode : {"replace", "zero", "average"}
        Operation to perform on the vxB axis. If "replace" (the default), the traces from the
        vxvxB component will be copied over to the vxB component. If "zero", the vxB component
        will be set to zero.

    Returns
    -------
    traces_vB_vvB : jax.Array
        the synthesised traces in vB x vvB axis (shower plane)
    """
    # here the threshold cut is rather high, I am not sure why such a high cut is necessary
    # but it is somehow necessary for the antenna set I am using
    ant_at_vvB_axis = jnp.abs(x) < 5
    ant_at_vB_axis = jnp.abs(y) < 5

    if vxB_axis_mode == "replace":
        traces_geoce = traces_geoce.at[:, ant_at_vB_axis, :].set(
            traces_geoce[:, ant_at_vvB_axis, :]
        )
    elif vxB_axis_mode == "zero":
        # for the vxB component, set to zero
        traces_geoce = traces_geoce.at[:, ant_at_vB_axis, :].set(0.0)
    elif vxB_axis_mode == "average":
        # for the vxB component, average the vxvxB and vxB components
        arm_idces = jnp.arange(0, len(x), 8, dtype=int)
        traces_geoce = traces_geoce.at[:, arm_idces, :].set(
            0.5
            * (traces_geoce[:, arm_idces + 1, :] + traces_geoce[:, arm_idces + 7, :])
        )
        traces_geoce = traces_geoce.at[:, arm_idces + 4, :].set(
            0.5
            * (traces_geoce[:, arm_idces + 3, :] + traces_geoce[:, arm_idces + 5, :])
        )

    else:
        raise ValueError(
            f"Unknown vxB_axis_mode: {vxB_axis_mode}. Must be one of 'replace' or 'zero'."
        )

    return traces_geoce


def reconstruct_starshape_from_vxvxB(
    geo_traces, ce_traces, antenna_positions_vvB, number_of_arms=8
):
    """
    Reconstruct traces on all arms from vxvxB-arm simulation (where geo and ce are orthogonal).

    This is used for generating star-shapes from simulations that only have antennas on the
    vxvxB arm. The traces on the other arms are simply copied from the vxvxB arm, and the
    antenna positions are rotated accordingly.

    Traces synthesised for all slices can also be passed. In that case, the output traces will
    have the same number of slices. If only at the ground, the n_slices = 1.

    Parameters
    ----------
    geo_traces : np.ndarray
        Geomagnetic traces on the vxvxB arm, shape (n_antennas, n_samples, n_slices)
    ce_traces : np.ndarray
        Charge-excess traces on the vxvxB arm, shape (n_antennas, n_samples, n_slices)
    antenna_positions_vvB : np.ndarray
        Positions of antennas on vxvxB arm (distance in shower plane), shape (n_antennas,)
    number_of_arms : int, default=8
        Total number of star-shape arms to reconstruct

    Returns
    -------
    geo_traces_starshape : np.ndarray
        Geomagnetic traces on all arms, shape (n_antennas*number_of_arms, n_samples, n_slices)
    ce_traces_starshape : np.ndarray
        Charge-excess traces on all arms, shape (n_antennas*number_of_arms, n_samples, n_slices)
    antenna_positions_starshape : np.ndarray
        Antenna positions on all arms, shape (n_antennas*number_of_arms, 3)
    """
    n_ant_per_arm = geo_traces.shape[0]

    # get the total number of antennas
    n_ant = n_ant_per_arm * number_of_arms

    # initialise arrays
    antenna_positions_starshape = jnp.zeros((n_ant, 3))
    geo_traces_starshape = jnp.zeros((n_ant, *geo_traces.shape[1:]))
    ce_traces_starshape = jnp.zeros((n_ant, *geo_traces.shape[1:]))
    antenna_labels = []

    phi_angles = jnp.linspace(0, 360, number_of_arms, endpoint=False)

    for ip, phi in enumerate(phi_angles):
        phi_rad = jnp.deg2rad(phi)
        geo_traces_starshape = geo_traces_starshape.at[
            ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, ...
        ].set(geo_traces)
        ce_traces_starshape = ce_traces_starshape.at[
            ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, ...
        ].set(ce_traces)
        antenna_positions_starshape = antenna_positions_starshape.at[
            ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, 0
        ].set(antenna_positions_vvB * jnp.cos(phi_rad))
        antenna_positions_starshape = antenna_positions_starshape.at[
            ip * n_ant_per_arm : (ip + 1) * n_ant_per_arm, 1
        ].set(antenna_positions_vvB * jnp.sin(phi_rad))

        antenna_labels += [f"ant_r{antenna_positions_vvB[i]:.0f}_phi{phi:.0f}" for i in range(n_ant_per_arm)]

    return geo_traces_starshape, ce_traces_starshape, antenna_positions_starshape, antenna_labels

def geo_ce_to_e(traces_geo: jax.typing.ArrayLike, traces_ce : jax.typing.ArrayLike, x, y) -> jax.Array:
    """
    Wrapper function for jax_radio_tools.trace_utils.geo_ce_to_vxB_vxvxB
    """
    return geo_ce_to_vB_vvB_v(
        jnp.array([traces_geo, traces_ce]),
        x,
        y
    )
